#  -*- coding: utf-8 -*-
# Copyright (C) Katelyn Breivik (2017 - 2021)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`Match`
"""

import numpy as np
import astropy.stats as astroStats
import warnings
from cosmic.utils import dat_transform


__author__ = "Katelyn Breivik <katie.breivik@gmail.com>"
__credits__ = "Scott Coughlin <scott.coughlin@ligo.org>"
__all__ = ["match", "perform_convergence"]


def match(dataCm):
    """Performs the Match calculation in Eq. 1 of Breivik & Larson (2018)

    Parameters
    ----------
    dataCm : list
        List of two cumulative data sets for a single paramter

    Returns
    -------
    match : list
        List of matches for each cumulative data set

    binwidth : float
        Binwidth of histograms used for match computation
    """

    # DEFINE A LIST TO HOLD THE BINNED DATA:
    histo = [[], []]
    histoBinEdges = [[], []]
    # COMPUTE THE BINWIDTH FOR THE MOST COMPLETE DATA SET:
    # NOTE: THIS WILL BE THE BINWIDTH FOR ALL THE HISTOGRAMS IN THE HISTO LIST
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message="divide by zero encountered in double_scalars"
        )
        try:
            bw, binEdges = astroStats.knuth_bin_width(
                np.array(dataCm[0]), return_bins=True
            )
        except Exception:
            bw, binEdges = astroStats.scott_bin_width(
                np.array(dataCm[0]), return_bins=True
            )
    if bw < 1e-4:
        bw = 1e-4
        binEdges = np.arange(binEdges[0], binEdges[-1], bw)

    # BIN THE DATA:
    for i in range(2):
        histo[i], histoBinEdges[i] = astroStats.histogram(
            dataCm[i], bins=binEdges, density=True
        )
    # COMPUTE THE MATCH:
    nominator = []
    denominator1 = []
    denominator2 = []
    nominatorSum = []
    denominator1Sum = []
    denominator2Sum = []

    histo2 = histo[1]
    histo1 = histo[0]

    for j in range(len(histo1)):
        nominator.append(histo1[j] * histo2[j])
        denominator1.append((histo1[j] * histo1[j]))
        denominator2.append((histo2[j] * histo2[j]))
    nominatorSum.append(np.sum(nominator))
    denominator1Sum.append(np.sum(denominator1))
    denominator2Sum.append(np.sum(denominator2))

    nominatorSum = np.array(nominatorSum)
    denominator1Sum = np.array(denominator1Sum)
    denominator2Sum = np.array(denominator2Sum)

    binwidth = binEdges[1] - binEdges[0]
    if binwidth < 1e-7:
        match = 1e-9
    else:
        match = np.log10(1 - nominatorSum / np.sqrt(denominator1Sum * denominator2Sum))

    return match[0], binwidth


def perform_convergence(conv_params, conv_1, conv_2, log_file):
    """Performs the convergence calculations for each convergence parameter
    and binary state

    Parameters
    ----------
    conv_params : dict
        List of user supplied convergence parameters
    conv_1 : DataFrame
        Cumulative data set of conv arrays
    conv_2 : DataFrame
        Most recent data set of conv array from most recent BSE run
    log_file : file write
        File to log matches or if the convergence params are not appropriate
        e.g. eccentricity for a disrupted system

    Returns
    -------
    match : array
        Matches for each cumulative data set
    """
    match_all = []
    for conv_param in conv_params:
        if (conv_param == "ecc") and (np.all(conv_1[conv_param] < 1e-7)):
            log_file.write(
                "{0} is circular or disrupted for all conv binaries\n".format(
                    conv_param
                )
            )
            match_all.append(-9)
        elif (conv_param == "ecc") and (np.all(conv_1[conv_param] == -1.0)):
            log_file.write(
                "{0} is the same for all disrupted binaries\n".format(conv_param)
            )
            match_all.append(-9)
        elif conv_param == "ecc":
            conv_1_ecc = conv_1.loc[conv_1.ecc > 0]
            if len(conv_1_ecc) < 10:
                log_file.write("not enough eccentric binaries to compute match")
                match_all.append(-9)
            else:
                conv_2_ecc = conv_2.loc[conv_2.ecc > 0]
                if len(conv_2_ecc) == len(conv_1_ecc):
                    conv_2_ecc = conv_2_ecc[: int(len(conv_2_ecc) / 2)]
                match_compute, bw = match(
                    [
                        dat_transform(conv_1_ecc, [conv_param])[0].tolist(),
                        dat_transform(conv_2_ecc, [conv_param])[0].tolist(),
                    ]
                )
                match_all.append(match_compute)

        elif (conv_param == "porb") and (
            (np.all(conv_1[conv_param] == 0.0)) or (np.all(conv_1[conv_param] == -1.0))
        ):
            log_file.write(
                "{0} is the same for all converging binaries\n".format(conv_param)
            )
            match_all.append(-9)
        elif (conv_param == "sep") and (
            (np.all(conv_1[conv_param] == 0.0)) or (np.all(conv_1[conv_param] == -1.0))
        ):
            log_file.write(
                "{0} is the same for all converging binaries\n".format(conv_param)
            )
            match_all.append(-9)
        else:
            if len(conv_2) == len(conv_1):
                conv_2 = conv_2[: int(len(conv_2) / 2)]
            match_compute, bw = match(
                [
                    dat_transform(conv_1, [conv_param])[0].tolist(),
                    dat_transform(conv_2, [conv_param])[0].tolist(),
                ]
            )
            match_all.append(match_compute)

    log_file.write("matches for converging population are are: {0}\n".format(match_all))
    log_file.write(
        "Number of binaries in converging population is: {0}\n".format(len(conv_1))
    )
    log_file.write("Binwidth is: {0}\n".format(bw))

    return match_all



from cosmic import _evolvebin
from .filter import parse_column_filters
import operator
import numpy

CHECKSTATE_COLUMNS = numpy.array(
    [
        "binstate",
        "evol_type",
        "mass_1",
        "mass_2",
        "kstar_1",
        "kstar_2",
        "sep",
        "porb",
        "ecc",
        "rrlo_1",
        "rrlo_2",
        "aj_1",
        "aj_2",
        "tms_1",
        "tms_2",
        "massc_1",
        "massc_2",
        "rad_1",
        "rad_2",
        "mass0_1",
        "mass0_2",
        "lum_1",
        "lum_2",
        "radc_1",
        "radc_2",
        "menv_1",
        "menv_2",
        "renv_1",
        "renv_2",
        "omega_spin_1",
        "omega_spin_2",
        "b0_1",
        "b0_2",
        "bacc_1",
        "bacc_2",
        "tacc_1",
        "tacc_2",
        "epoch_1",
        "epoch_2",
        "bhspin_1",
        "bhspin_2",
    ]
)

DEFAULT_CONDITIONS = [-10e30, -1, 10e30] * CHECKSTATE_COLUMNS.size
DEFAULT_CONDITIONS = numpy.array([DEFAULT_CONDITIONS] * 15)
DEFAULT_DTP_STATE = -1 * numpy.ones(15)


def set_checkstates(timestep_conditions=[]):
    """A function which will detemine different time resolution for different states

    Parameters:
        timestep_conditions : (list) default empty (no dynamic dtp setting)
            a nested list of the many different time resolutions and conditions
            for which you would like to apply those time resolution, e.g.,

            >>> timestep_conditions = [['20.0<mass_1<25.5', '15.5>mass_2>10.0', 'dtp=1.0'],
            >>> ['kstar_1=14', 'lum_1>10.0', 'dtp=0.01'],
            >>> ['2>=binstate>=1', 'dtp=None']]

            The first two sets of conditons would be done in a AND fashion.
            The last condition would end time resolution for the BCM array and it
            would skip to only printing the final state
    """
    # assume that we are not doing any special dtp setting
    _evolvebin.check_dtp.check_dtp = 0

    # Set the default state for checkstate which is that there are no
    # conditional states at which to set a special dtp
    checkstate_array = getattr(_evolvebin.checkstate_array, "checkstate_array")
    checkstate_array[:, :] = DEFAULT_CONDITIONS

    # Again we assume that no condtions exist to set a special dtp
    dtp_state = getattr(_evolvebin.checkstate_params, "dtp_state")
    dtp_state[:] = DEFAULT_DTP_STATE

    for index, condition in enumerate(timestep_conditions):
        # we are checking for conditions
        _evolvebin.check_dtp.check_dtp = 1
        conditions = parse_column_filters(condition)
        for param in conditions:
            # find where in the checkstate_array this param is
            param_index = numpy.argwhere(param[0].lower() == CHECKSTATE_COLUMNS)
            if param[0] == "dtp":
                dtp_state = getattr(_evolvebin.checkstate_params, "dtp_state")
                if param[2] == "None":
                    dtp_state[index] = 13700.0
                else:
                    dtp_state[index] = param[2]
                continue

            if param[1] == operator.eq:
                checkstate_array[index, param_index * 3] = param[2]
                checkstate_array[index, param_index * 3 + 2] = param[2]
                checkstate_array[index, param_index * 3 + 1] = 0
            elif param[1] == operator.gt:
                checkstate_array[index, param_index * 3] = param[2]
                checkstate_array[index, param_index * 3 + 1] = 1
            elif param[1] == operator.ge:
                checkstate_array[index, param_index * 3] = param[2]
                checkstate_array[index, param_index * 3 + 1] = 2
            elif param[1] == operator.lt:
                checkstate_array[index, param_index * 3 + 2] = param[2]
                checkstate_array[index, param_index * 3 + 1] = 3
            elif param[1] == operator.le:
                checkstate_array[index, param_index * 3 + 2] = param[2]
                checkstate_array[index, param_index * 3 + 1] = 4



# -*- coding: utf-8 -*-
# Copyright (C) Katelyn Breivik (2017 - 2021)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`cmc`
"""

import numpy as np

from .sampler import register_sampler
from .independent import Sample
from .. import InitialCMCTable, InitialBinaryTable
from ..cmc import elson, king
from ... import utils

__author__ = "Scott Coughlin <scott.coughlin@ligo.org>"
__credits__ = [
   "Carl Rodriguez <carllouisrodriguez@gmail.com>", 
   "Newlin Weatherford <newlinweatherford2017@u.northwestern.edu>"
]
__all__ = ["get_cmc_sampler", "CMCSample"]


def get_cmc_sampler(
    cluster_profile, primary_model, ecc_model, porb_model, binfrac_model, met, size, **kwargs
):
    """Generates an initial binary sample according to user specified models

    Parameters
    ----------
    cluster_profile : `str`
        Model to use for the cluster profile (i.e. sampling of the placement of objects in the cluster and their velocity within the cluster)
        Options include:
        'plummer' : Standard Plummer sphere.
            Additional parameters:
            'r_max' : `float`
                the maximum radius (in virial radii) to sample the clsuter
        'elson' : EFF 1987 profile.  Generalization of Plummer that better fits young massive clusters
            Additional parameters:
            'gamma' : `float`
                steepness paramter for Elson profile; note that gamma=4 is same is Plummer
            'r_max' : `float`
                the maximum radius (in virial radii) to sample the clsuter
        'king' : King profile 
            'w_0' : `float`
                King concentration parameter
            'r_max' : `float`
                the maximum radius (in virial radii) to sample the clsuter

    primary_model : `str`
        Model to sample primary mass; choices include: kroupa93, kroupa01, salpeter55, custom
        if 'custom' is selected, must also pass arguemts:
        alphas : `array`
            list of power law indicies
        mcuts : `array`
            breaks in the power laws.
        e.g. alphas=[-1.3,-2.3,-2.3],mcuts=[0.08,0.5,1.0,150.] reproduces standard Kroupa2001 IMF

    ecc_model : `str`
        Model to sample eccentricity; choices include: thermal, uniform, sana12

    porb_model : `str`
        Model to sample orbital period; choices include: log_uniform, sana12

    msort : `float`
        Stars with M>msort can have different pairing and sampling of companions

    pair : `float`
        Sets the pairing of stars M>msort only with stars with M>msort

    binfrac_model : `str or float`
        Model for binary fraction; choices include: vanHaaften or a fraction where 1.0 is 100% binaries

    binfrac_model_msort : `str or float`
        Same as binfrac_model for M>msort

    qmin : `float`
        kwarg which sets the minimum mass ratio for sampling the secondary
        where the mass ratio distribution is flat in q
        if q > 0, qmin sets the minimum mass ratio
        q = -1, this limits the minimum mass ratio to be set such that
        the pre-MS lifetime of the secondary is not longer than the full
        lifetime of the primary if it were to evolve as a single star

    m2_min : `float`
        kwarg which sets the minimum secondary mass for sampling
        the secondary as uniform in mass_2 between m2_min and mass_1

    qmin_msort : `float`
        Same as qmin for M>msort; only applies if qmin is supplied

    met : `float`
        Sets the metallicity of the binary population where solar metallicity is zsun 

    size : `int`
        Size of the population to sample

    zsun : `float`
        optional kwarg for setting effective radii, default is 0.02

    Optional Parameters
    -------------------
    virial_radius : `float`
        the initial virial radius of the cluster, in parsecs
        Default -- 1 pc

    tidal_radius : `float`
        the initial tidal radius of the cluster, in units of the virial_radius
        Default -- 1e6 rvir

    seed : `float`
        seed to the random number generator, for reproducability

    Returns
    -------
    Singles: `pandas.DataFrame`
        DataFrame of Single objects in the format of the InitialCMCTable

    Binaries: `pandas.DataFrame`
        DataFrame of Single objects in the format of the InitialCMCTable
    """
    
    initconditions = CMCSample()

    # if RNG seed is provided, then use it globally
    rng_seed = kwargs.pop("seed", 0)
    if rng_seed != 0:
        np.random.seed(rng_seed)

    # get radii, radial and transverse velocities
    r, vr, vt = initconditions.set_r_vr_vt(cluster_profile, N=size, **kwargs)

    # track the mass in singles and the mass in binaries
    mass_singles = 0.0
    mass_binaries = 0.0

    # track the total number of stars sampled
    n_singles = 0
    n_binaries = 0

    mass1, total_mass1 = initconditions.sample_primary(
        primary_model, size=size, **kwargs)
    (
        mass1_binaries,
        mass_single,
        binfrac_binaries,
        binary_index,
    ) = initconditions.binary_select(mass1, binfrac_model=binfrac_model, **kwargs)

    mass2_binaries = initconditions.sample_secondary(
        mass1_binaries, **kwargs)

    # track the mass sampled
    mass_singles += np.sum(mass_single)
    mass_binaries += np.sum(mass1_binaries)
    mass_binaries += np.sum(mass2_binaries)

    # set singles id
    single_ids = np.arange(mass1.size)
    binary_secondary_object_id = np.arange(mass1.size, mass1.size + mass2_binaries.size)

    # set binind and correct masses of binaries
    binind = np.zeros(mass1.size)
    binind[binary_index] = np.arange(len(binary_index)) + 1
    mass1[binary_index] += mass2_binaries

    # track the total number sampled
    n_singles += len(mass_single)
    n_binaries += len(mass1_binaries)

    # Obtain radii (technically this is done for the binaries in the independent sampler 
    # if set_radii_with_BSE is true, but that's not a huge amount of overhead)
    zsun = kwargs.pop("zsun", 0.02)

    Reff = initconditions.set_reff(mass1, metallicity=met, zsun=zsun)
    Reff1 = Reff[binary_index]
    Reff2 = initconditions.set_reff(mass2_binaries, metallicity=met, zsun=zsun)

    # select out the primaries and secondaries that will produce the final kstars
    porb_max = initconditions.calc_porb_max(mass1, vr, vt, binary_index, mass1_binaries, mass2_binaries, **kwargs)

    porb,aRL_over_a = initconditions.sample_porb(
        mass1_binaries, mass2_binaries, Reff1, Reff2, porb_model=porb_model, porb_max=porb_max, size=mass1_binaries.size
    )
    ecc = initconditions.sample_ecc(aRL_over_a, ecc_model, size=mass1_binaries.size)

    sep = utils.a_from_p(porb, mass1_binaries, mass2_binaries)
    kstar1 = initconditions.set_kstar(mass1_binaries)
    kstar2 = initconditions.set_kstar(mass2_binaries)

    singles_table = InitialCMCTable.InitialCMCSingles(
        single_ids +
        1, initconditions.set_kstar(mass1), mass1, Reff, r, vr, vt, binind
    )

    binaries_table = InitialCMCTable.InitialCMCBinaries(
        np.arange(mass1_binaries.size) + 1,
        single_ids[binary_index] + 1,
        kstar1,
        mass1_binaries,
        Reff1,
        binary_secondary_object_id + 1,
        kstar2,
        mass2_binaries,
        Reff2,
        sep,
        ecc,
    )

    singles_table.metallicity = met
    singles_table.mass_of_cluster = np.sum(singles_table["m"])
    binaries_table.metallicity = met
    singles_table.virial_radius = kwargs.get("virial_radius",1) 
    singles_table.tidal_radius = kwargs.get("tidal_radius",1e6) 

    return singles_table, binaries_table


register_sampler(
    "cmc",
    InitialCMCTable,
    get_cmc_sampler,
    usage="primary_model ecc_model porb_model binfrac_model met size",
)

def get_cmc_point_mass_sampler(
    cluster_profile, size, **kwargs
):
    """Generates an CMC cluster model according  

    Parameters
    ----------
    cluster_profile : `str`
        Model to use for the cluster profile (i.e. sampling of the placement of objects in the cluster and their velocity within the cluster)
        Options include:
        'plummer' : Standard Plummer sphere.
            Additional parameters:
            'r_max' : `float`
                the maximum radius (in virial radii) to sample the clsuter
        'elson' : EFF 1987 profile.  Generalization of Plummer that better fits young massive clusters
            Additional parameters:
            'gamma' : `float`
                steepness paramter for Elson profile; note that gamma=4 is same is Plummer
            'r_max' : `float`
                the maximum radius (in virial radii) to sample the clsuter
        'king' : King profile 
            'w_0' : `float`
                King concentration parameter
            'r_max' : `float`
                the maximum radius (in virial radii) to sample the clsuter

    size : `int`
        Size of the population to sample

    Returns
    -------
    Singles: `pandas.DataFrame`
        DataFrame of Single objects in the format of the InitialCMCTable
    Binaries: `pandas.DataFrame`
        DataFrame of Single objects in the format of the InitialCMCTable

    """

    initconditions = CMCSample()

    # get radii, radial and transverse velocities
    r, vr, vt = initconditions.set_r_vr_vt(cluster_profile, N=size, **kwargs)

    mass1 = np.ones(size)/size
    Reff = np.zeros(size)
    binind = np.zeros(size)
    single_ids = np.arange(size)

    singles_table = InitialCMCTable.InitialCMCSingles(
        single_ids + 1, initconditions.set_kstar(mass1), mass1, Reff, r, vr, vt, binind
    )
    # Already scaled from the IC generators
    singles_table.scaled_to_nbody_units = True

    # We assume no binaries for the point mass models
    binaries_table = InitialCMCTable.InitialCMCBinaries(1,1,0,0,0,1,0,0,0,0,0)

    singles_table.metallicity = 0.02
    singles_table.mass_of_cluster = np.sum(singles_table["m"])*size
    binaries_table.metallicity = 0.02
    singles_table.virial_radius = kwargs.get("virial_radius",1) 
    singles_table.tidal_radius = kwargs.get("tidal_radius",1e6) 
   
    return singles_table, binaries_table

register_sampler(
    "cmc_point_mass",
    InitialCMCTable,
    get_cmc_point_mass_sampler,
    usage="size",
)

class CMCSample(Sample):
    def set_r_vr_vt(self,cluster_profile, **kwargs):
        if cluster_profile == "elson":
            elson_kwargs = {
                k: v for k, v in kwargs.items() if k in ["gamma", "r_max", "N"]
            }
            r, vr, vt = elson.draw_r_vr_vt(**elson_kwargs)
        elif cluster_profile == "plummer":
            plummer_kwargs = {k: v for k, v in kwargs.items() if k in ["r_max", "N"]}
            r, vr, vt = elson.draw_r_vr_vt(gamma=4, **plummer_kwargs)
        elif cluster_profile == "king":
            king_kwargs = {k: v for k, v in kwargs.items() if k in ["w_0", "N"]}
            r, vr, vt = king.draw_r_vr_vt(**king_kwargs)
        else:
            raise ValueError("Cluster profile not defined! Please specify one of [elson, plummer, king]")

        return r, vr, vt

    def calc_porb_max(self, mass, vr, vt, binary_index, mass1_binary, mass2_binary, **kwargs): 

        ## First, compute a rolling average with window length of AVEKERNEL
        ## NOTE: this is in cluster code units (i.e. M=G=1), same as vr and vt
        AVEKERNEL = 20

        ## Then compute the average mass and avergae of m*v^2
        ## First, to do this properly, we need to reflect the boundary points (so that the average 
        ## at the boundary doesn't go to zero artifically))
        m = np.concatenate([mass[AVEKERNEL-1::-1],mass,mass[:-AVEKERNEL-1:-1]])
        mv2 = mass*(vr*vr + vt*vt) 
        mv2 = np.concatenate([mv2[AVEKERNEL-1::-1],mv2,mv2[:-AVEKERNEL-1:-1]])

        ## Then compute the averages by convolving with a uniform array (note we trim the reflected points off here)
        m_average = np.convolve(m,np.ones(AVEKERNEL),mode='same')[AVEKERNEL:-AVEKERNEL]/AVEKERNEL
        mv2_average = np.convolve(mv2,np.ones(AVEKERNEL),mode='same')[AVEKERNEL:-AVEKERNEL]/AVEKERNEL

        ## Finally return the mass-weighted average 3D velocity dispersion (in cluster code units)
        sigma =  np.sqrt(mv2_average/m_average)

        ## Now compute the orbital velocity corresponding to the hard/soft boundary; we only want it for the binaries
        v_orb = 0.7*1.30294*sigma[binary_index]   #sigma * 4/sqrt(3/pi); 0.7 is a factor we use in CMC
        #v_orb = 0.7*1.30294*np.mean(sigma[:20])   #old CMC way of using just core velocity dispersion; TODO: possibly make flag option?


        ## Maximum semi-major axis just comes from Kepler's 3rd
        ## Note, to keep in code units, we need to divide binary masses by total cluster mass
        amax = (mass1_binary+mass2_binary) / v_orb**2 / np.sum(mass) 

        ## Convert from code units (virial radii) to RSUN 
        virial_radius = kwargs.get("virial_radius",1) ## get the virial radius of the cluster (uses 1pc if not given)
        RSUN_PER_PARSEC = 4.435e+7
        amax *= RSUN_PER_PARSEC * virial_radius 

        ## Finally go from sep to porb
        porb_max = utils.p_from_a(amax, mass1_binary, mass2_binary)

        return porb_max ## returns orbital period IN DAYS





# -*- coding: utf-8 -*-
# Copyright (C) Scott Coughlin (2017 - 2021)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`InitialBinaryTable`
"""

import numpy as np
import math
import sys

import pandas as pd

__author__ = "Katelyn Breivik <katie.breivik@gmail.com>"
__credits__ = "Scott Coughlin <scott.coughlin@ligo.org>"
__all__ = ["InitialBinaryTable"]


G = 6.67384 * math.pow(10, -11.0)
c = 2.99792458 * math.pow(10, 8.0)
parsec = 3.08567758 * math.pow(10, 16)
Rsun = 6.955 * math.pow(10, 8)
Msun = 1.9891 * math.pow(10, 30)
day = 86400.0
rsun_in_au = 215.0954
day_in_year = 365.242
sec_in_day = 86400.0
sec_in_hour = 3600.0
hrs_in_day = 24.0
sec_in_year = 3.15569 * 10 ** 7.0
Tobs = 3.15569 * 10 ** 7.0
geo_mass = G / c ** 2

INITIAL_CONDITIONS_COLUMNS = []

INITIAL_CONDITIONS_COLUMNS_CORE = [
    "kstar_1",
    "kstar_2",
    "mass_1",
    "mass_2",
    "porb",
    "ecc",
    "metallicity",
    "tphysf",
]

INITIAL_CONDITIONS_COLUMNS_EXTRAS = [
    "mass0_1",
    "mass0_2",
    "rad_1",
    "rad_2",
    "lum_1",
    "lum_2",
    "massc_1",
    "massc_2",
    "radc_1",
    "radc_2",
    "menv_1",
    "menv_2",
    "renv_1",
    "renv_2",
    "omega_spin_1",
    "omega_spin_2",
    "B_1",
    "B_2",
    "bacc_1",
    "bacc_2",
    "tacc_1",
    "tacc_2",
    "epoch_1",
    "epoch_2",
    "tms_1",
    "tms_2",
    "bhspin_1",
    "bhspin_2",
    "tphys",
]

INITIAL_CONDITIONS_COLUMNS.extend(INITIAL_CONDITIONS_COLUMNS_CORE)
INITIAL_CONDITIONS_COLUMNS.extend(INITIAL_CONDITIONS_COLUMNS_EXTRAS)

INITIAL_CONDITIONS_MISC_COLUMNS = ["binfrac"]

if sys.version_info.major == 2 and sys.version_info.minor == 7:
    INITIAL_CONDITIONS_COLUMNS_ALL = INITIAL_CONDITIONS_COLUMNS[:]
else:
    INITIAL_CONDITIONS_COLUMNS_ALL = INITIAL_CONDITIONS_COLUMNS.copy()

INITIAL_CONDITIONS_COLUMNS_ALL.extend(INITIAL_CONDITIONS_MISC_COLUMNS)


class InitialBinaryTable:
    @classmethod
    def InitialBinaries(
        cls, m1, m2, porb, ecc, tphysf, kstar1, kstar2, metallicity, **kwargs
    ):
        """Create single binary

        Parameters
        ----------
        m1 : float
            Primary mass [Msun]
        m2 : float
            Secondary mass [Msun]
        porb : float
            Orbital period [days]
        ecc : float
            Eccentricity
        tphysf : float
            Time to evolve the binary [Myr]
        kstar1 : array
            0-14 Initial stellar type of the larger object;
            main sequence stars are 0 if m < 0.7 Msun and 1 otherwise
        kstar2 : array
            0-14 Initial stellar type of the smaller object;
            main sequence stars are 0 if m < 0.7 Msun and 1 otherwise
        metallicity : float
            Metallicity of the binaries; Z_sun = 0.02

        **kwargs

            binfrac : float
                System-specific probability of the primary star being in a binary

            mass0_1,mass0_2,rad1,rad2,lumin1,lumin2,
            massc1,massc2,radc1,radc2,menv1,menv2,renv1,renv2,
            ospin1,ospin2,b_0_1,b_0_2,bacc1,bacc2,
            tacc1,tacc2,epoch1,epoch2,tms1,tms2
            bhspin1,bhspin2

        Returns
        -------
        InitialBinaries : DataFrame
            Single binary initial conditions

        """
        # System-specific probability of the primary star being in a binary
        binfrac = kwargs.pop("binfrac", np.ones(np.array(m1).size))

        # The following are in general not needed for COSMIC, however
        # if you wish to start a binary at a very specific point in its evolution
        # these kwarg arguments can help you do this.
        # For instance the Globular Cluster code CMC requires this behavior.
        mass0_1 = kwargs.pop("mass0_1", m1)
        mass0_2 = kwargs.pop("mass0_2", m2)
        rad1 = kwargs.pop("rad_1", np.zeros(np.array(m1).size))
        rad2 = kwargs.pop("rad_2", np.zeros(np.array(m1).size))
        lumin1 = kwargs.pop("lumin_1", np.zeros(np.array(m1).size))
        lumin2 = kwargs.pop("lumin_2", np.zeros(np.array(m1).size))
        massc1 = kwargs.pop("massc_1", np.zeros(np.array(m1).size))
        massc2 = kwargs.pop("massc_2", np.zeros(np.array(m1).size))
        radc1 = kwargs.pop("radc_1", np.zeros(np.array(m1).size))
        radc2 = kwargs.pop("radc_2", np.zeros(np.array(m1).size))
        menv1 = kwargs.pop("menv_1", np.zeros(np.array(m1).size))
        menv2 = kwargs.pop("menv_2", np.zeros(np.array(m1).size))
        renv1 = kwargs.pop("renv_1", np.zeros(np.array(m1).size))
        renv2 = kwargs.pop("renv_2", np.zeros(np.array(m1).size))
        ospin1 = kwargs.pop("ospin_1", np.zeros(np.array(m1).size))
        ospin2 = kwargs.pop("ospin_2", np.zeros(np.array(m1).size))
        b_0_1 = kwargs.pop("B_1", np.zeros(np.array(m1).size))
        b_0_2 = kwargs.pop("B_2", np.zeros(np.array(m1).size))
        bacc1 = kwargs.pop("bacc_1", np.zeros(np.array(m1).size))
        bacc2 = kwargs.pop("bacc_2", np.zeros(np.array(m1).size))
        tacc1 = kwargs.pop("tacc_1", np.zeros(np.array(m1).size))
        tacc2 = kwargs.pop("tacc_2", np.zeros(np.array(m1).size))
        epoch1 = kwargs.pop("epoch_1", np.zeros(np.array(m1).size))
        epoch2 = kwargs.pop("epoch_2", np.zeros(np.array(m1).size))
        tms1 = kwargs.pop("tms_1", np.zeros(np.array(m1).size))
        tms2 = kwargs.pop("tms_2", np.zeros(np.array(m1).size))
        bhspin1 = kwargs.pop("bhspin_1", np.zeros(np.array(m1).size))
        bhspin2 = kwargs.pop("bhspin_2", np.zeros(np.array(m1).size))
        tphys = kwargs.pop("tphys", np.zeros(np.array(m1).size))

        bin_dat = pd.DataFrame(
            np.vstack(
                [
                    kstar1,
                    kstar2,
                    m1,
                    m2,
                    porb,
                    ecc,
                    metallicity,
                    tphysf,
                    mass0_1,
                    mass0_2,
                    rad1,
                    rad2,
                    lumin1,
                    lumin2,
                    massc1,
                    massc2,
                    radc1,
                    radc2,
                    menv1,
                    menv2,
                    renv1,
                    renv2,
                    ospin1,
                    ospin2,
                    b_0_1,
                    b_0_2,
                    bacc1,
                    bacc2,
                    tacc1,
                    tacc2,
                    epoch1,
                    epoch2,
                    tms1,
                    tms2,
                    bhspin1,
                    bhspin2,
                    tphys,
                    binfrac,
                ]
            ).T,
            columns=INITIAL_CONDITIONS_COLUMNS_ALL,
        )

        return bin_dat

    @classmethod
    def sampler(cls, format_, *args, **kwargs):
        """Fetch a method to generate an initial binary sample

        Parameters
        ----------
        format : str
            the method name; Choose from 'independent' or 'multidim'

        *args
            the arguments necessary for the registered sample
            method; see help(InitialBinaryTable.sampler('independent')
            to see the arguments necessary for the independent sample
        """
        # standard registered fetch
        from .sampler.sampler import get_sampler

        sampler = get_sampler(format_, cls)
        return sampler(*args, **kwargs)



# -*- coding: utf-8 -*-
# Copyright (C) Katelyn Breivik (2017 - 2021)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`independent`
"""

import numpy as np
import warnings
import pandas as pd

from ... import utils

from .sampler import register_sampler
from .. import InitialBinaryTable
from ... import _evolvebin


__author__ = "Katelyn Breivik <katie.breivik@gmail.com>"
__credits__ = ("Scott Coughlin <scott.coughlin@ligo.org>, Michael Zevin <michael.j.zevin@gmail.com>, "
               "Tom Wagg <tomjwagg@gmail.com>")
__all__ = ["get_independent_sampler", "Sample"]


def get_independent_sampler(
    final_kstar1,
    final_kstar2,
    primary_model,
    ecc_model,
    porb_model,
    SF_start,
    SF_duration,
    binfrac_model,
    met,
    size=None,
    total_mass=np.inf,
    sampling_target="size",
    trim_extra_samples=False,
    **kwargs
):
    """Generates an initial binary sample according to user specified models

    Parameters
    ----------
    final_kstar1 : `int or list`
        Int or list of final kstar1

    final_kstar2 : `int or list`
        Int or list of final kstar2

    primary_model : `str`
        Model to sample primary mass; choices include: kroupa93, kroupa01, salpeter55, custom
        if 'custom' is selected, must also pass arguemts:
        alphas : `array`
            list of power law indicies
        mcuts : `array`
            breaks in the power laws.
        e.g. alphas=[-1.3,-2.3,-2.3],mcuts=[0.08,0.5,1.0,150.] reproduces standard Kroupa2001 IMF

    ecc_model : `str`
        Model to sample eccentricity; choices include: thermal, uniform, sana12

    porb_model : `str`
        Model to sample orbital period; choices include: log_uniform, sana12, raghavan10, moe19

    qmin : `float`
        kwarg which sets the minimum mass ratio for sampling the secondary
        where the mass ratio distribution is flat in q
        if q > 0, qmin sets the minimum mass ratio
        q = -1, this limits the minimum mass ratio to be set such that
        the pre-MS lifetime of the secondary is not longer than the full
        lifetime of the primary if it were to evolve as a single star

    m_max : `float`
        kwarg which sets the maximum primary and secondary mass for sampling
        NOTE: this value changes the range of the IMF and should *not* be used
            as a means of selecting certain kstar types!

    m1_min : `float`
        kwarg which sets the minimum primary mass for sampling
        NOTE: this value changes the range of the IMF and should *not* be used
            as a means of selecting certain kstar types!

    m2_min : `float`
        kwarg which sets the minimum secondary mass for sampling
        the secondary as uniform in mass_2 between m2_min and mass_1

    msort : `float`
        Stars with M>msort can have different pairing and sampling of companions

    qmin_msort : `float`
        Same as qmin for M>msort

    m2_min_msort : `float`
        Same as m2_min for M>msort

    SF_start : `float`
        Time in the past when star formation initiates in Myr

    SF_duration : `float`
        Duration of constant star formation beginning from SF_Start in Myr

    binfrac_model : `str or float`
        Model for binary fraction; choices include: vanHaaften or a fraction where 1.0 is 100% binaries

    binfrac_model_msort : `str or float`
        Same as binfrac_model for M>msort

    met : `float`
        Sets the metallicity of the binary population where solar metallicity is zsun

    size : `int`
        Size of the population to sample

    total_mass : `float`
        Total mass to use as a target for sampling

    sampling_target : `str`
        Which type of target to use for sampling (either "size" or "total_mass"), by default "size".
        Note that `total_mass` must not be None when `sampling_target=="total_mass"`.

    trim_extra_samples : `str`
        Whether to trim the sampled population so that the total mass sampled is as close as possible to
        `total_mass`. Ignored when `sampling_target==size`.
        Note that given the discrete mass of stars, this could mean your sample is off by 300
        solar masses in the worst case scenario (of a 150+150 binary being sampled). In reality the majority
        of cases track the target total mass to within a solar mass.        

    zsun : `float`
        optional kwarg for setting effective radii, default is 0.02

    Returns
    -------
    InitialBinaryTable : `pandas.DataFrame`
        DataFrame in the format of the InitialBinaryTable

    mass_singles : `float`
        Total mass in single stars needed to generate population

    mass_binaries : `float`
        Total mass in binaries needed to generate population

    n_singles : `int`
        Number of single stars needed to generate a population

    n_binaries : `int`
        Number of binaries needed to generate a population
    """
    if sampling_target == "total_mass" and (total_mass is None or total_mass == np.inf):
        raise ValueError("If `sampling_target == 'total mass'` then `total_mass` must be supplied")
    if size is None and (total_mass is None or total_mass == np.inf):
        raise ValueError("Either a sample `size` or `total_mass` must be supplied")
    elif size is None:
        size = int(total_mass)

    if binfrac_model == 0.0 and sampling_target == "size":
        raise ValueError(("If `binfrac_model == 0.0` then `sampling_target` must be 'total_mass'. Otherwise "
                          "you are targetting a population of `size` binaries but will never select any."))

    final_kstar1 = [final_kstar1] if isinstance(final_kstar1, (int, float)) else final_kstar1
    final_kstar2 = [final_kstar2] if isinstance(final_kstar2, (int, float)) else final_kstar2
    primary_min, primary_max, secondary_min, secondary_max = utils.mass_min_max_select(
        final_kstar1, final_kstar2, **kwargs)
    initconditions = Sample()

    # set up multiplier if the mass sampling is inefficient
    multiplier = 1

    # track samples to actually return (after masks)
    mass1_singles = []
    mass1_binary = []
    mass2_binary = []
    binfrac = []

    # track the total mass of singles and binaries sampled
    m_sampled_singles = 0.0
    m_sampled_binaries = 0.0

    # track the total number of stars sampled
    n_singles = 0
    n_binaries = 0

    # if porb_model = `moe19`, the binary fraction is fixed based on the metallicity 
    if porb_model == "moe19":
        binfrac_model = utils.get_met_dep_binfrac(met)
        warnings.warn('your supplied binfrac_model has been overwritten to {} match Moe+2019'.format(binfrac_model))

    # define a function that evaluates whether you've reached your sampling target
    target = lambda mass1_binary, size, m_sampled_singles, m_sampled_binaries, total_mass:\
        len(mass1_binary) < size if sampling_target == "size" else m_sampled_singles + m_sampled_binaries < total_mass

    # sample until you've reached your target
    while target(mass1_binary, size, m_sampled_singles, m_sampled_binaries, total_mass):
        # sample primary masses
        mass1, _ = initconditions.sample_primary(primary_model, size=int(size * multiplier), **kwargs)

        # split them into binaries or single stars
        (mass1_binaries, mass_single, binfrac_binaries, binary_index,
        ) = initconditions.binary_select(mass1, binfrac_model=binfrac_model, **kwargs)

        # sample secondary masses for the single stars
        mass2_binaries = initconditions.sample_secondary(mass1_binaries, **kwargs)

        # check if this batch of samples will take us over our sampling target
        if not target(mass1_binary, size,
                      m_sampled_singles + np.sum(mass_single),
                      m_sampled_binaries + np.sum(mass1_binaries) + np.sum(mass2_binaries),
                      total_mass) and trim_extra_samples and sampling_target == "total_mass":
            # get the cumulative total mass of the samples
            total_mass_list = np.copy(mass1)
            total_mass_list[binary_index] += mass2_binaries
            sampled_so_far = m_sampled_singles + m_sampled_binaries
            cumulative_total_mass = sampled_so_far + np.cumsum(total_mass_list)

            # find the boundary for reaching the right total mass
            threshold_index = np.where(cumulative_total_mass > total_mass)[0][0]

            keep_offset = abs(cumulative_total_mass[threshold_index] - total_mass)
            drop_offset = abs(cumulative_total_mass[threshold_index - 1] - total_mass)
            lim = threshold_index - 1 if (keep_offset > drop_offset) else threshold_index
            
            # work out how many singles vs. binaries to delete
            one_if_binary = np.zeros(len(mass1))
            one_if_binary[binary_index] = 1
            sb_delete = one_if_binary[lim + 1:]
            n_single_delete = (sb_delete == 0).sum()
            n_binary_delete = (sb_delete == 1).sum()

            # delete em!
            if n_single_delete > 0:
                mass_single = mass_single[:-n_single_delete]
            if n_binary_delete > 0:
                mass1_binaries = mass1_binaries[:-n_binary_delete]
                mass2_binaries = mass2_binaries[:-n_binary_delete]
                binfrac_binaries = binfrac_binaries[:-n_binary_delete]

            # ensure we don't loop again after this
            target = lambda mass1_binary, size, m_sampled_singles, m_sampled_binaries, total_mass: False

        # track the mass sampled
        m_sampled_singles += sum(mass_single)
        m_sampled_binaries += sum(mass1_binaries)
        m_sampled_binaries += sum(mass2_binaries)

        # track the total number sampled
        n_singles += len(mass_single)
        n_binaries += len(mass1_binaries)

        # select out the primaries and secondaries that will produce the final kstars
        ind_select = (  (mass1_binaries > primary_min)
                      & (mass1_binaries < primary_max)
                      & (mass2_binaries > secondary_min)
                      & (mass2_binaries < secondary_max))
        mass1_binary.extend(mass1_binaries[ind_select])
        mass2_binary.extend(mass2_binaries[ind_select])
        binfrac.extend(binfrac_binaries[ind_select])

        # select out the single stars that will produce the final kstar
        mass1_singles.extend(mass_single[(mass_single > primary_min) & (mass_single < primary_max)])

        # check to see if we should increase the multiplier factor to sample the population more quickly
        if target(mass1_binary, size / 100, m_sampled_singles, m_sampled_binaries, total_mass / 100):
            # well this sampling rate is clearly not working time to increase
            # the multiplier by an order of magnitude
            multiplier *= 10

    mass1_binary = np.array(mass1_binary)
    mass2_binary = np.array(mass2_binary)
    binfrac = np.asarray(binfrac)
    mass1_singles = np.asarray(mass1_singles)

    zsun = kwargs.pop("zsun", 0.02)

    rad1 = initconditions.set_reff(mass1_binary, metallicity=met, zsun=zsun)
    rad2 = initconditions.set_reff(mass2_binary, metallicity=met, zsun=zsun)

    # sample periods and eccentricities
    # if the porb_model is moe19, the metallicity needs to be supplied
    if porb_model == "moe19":
        porb,aRL_over_a = initconditions.sample_porb(
            mass1_binary, mass2_binary, rad1, rad2, porb_model, met=met, size=mass1_binary.size
        )
    else:
        porb,aRL_over_a = initconditions.sample_porb(
            mass1_binary, mass2_binary, rad1, rad2, porb_model, size=mass1_binary.size
        )
    ecc = initconditions.sample_ecc(aRL_over_a, ecc_model, size=mass1_binary.size)

    tphysf, metallicity = initconditions.sample_SFH(
        SF_start=SF_start, SF_duration=SF_duration, met=met, size=mass1_binary.size
    )
    metallicity[metallicity < 1e-4] = 1e-4
    metallicity[metallicity > 0.03] = 0.03
    kstar1 = initconditions.set_kstar(mass1_binary)
    kstar2 = initconditions.set_kstar(mass2_binary)

    if kwargs.pop("keep_singles", False):
        binary_table = InitialBinaryTable.InitialBinaries(
            mass1_binary,
            mass2_binary,
            porb,
            ecc,
            tphysf,
            kstar1,
            kstar2,
            metallicity,
            binfrac=binfrac,
        )
        tphysf_singles, metallicity_singles = initconditions.sample_SFH(
            SF_start=SF_start, SF_duration=SF_duration, met=met, size=mass1_singles.size
        )
        metallicity_singles[metallicity_singles < 1e-4] = 1e-4
        metallicity_singles[metallicity_singles > 0.03] = 0.03
        kstar1_singles = initconditions.set_kstar(mass1_singles)
        singles_table = InitialBinaryTable.InitialBinaries(
            mass1_singles,                          # mass1
            np.ones_like(mass1_singles) * 0,        # mass2 (all massless remnants)
            np.ones_like(mass1_singles) * -1,       # porb (single not binary)
            np.ones_like(mass1_singles) * -1,       # ecc (single not binary)
            tphysf_singles,                         # tphysf
            kstar1_singles,                         # kstar1
            np.ones_like(mass1_singles) * 15,       # kstar2 (all massless remnants)
            metallicity_singles,                    # metallicity
        )
        binary_table = pd.concat([binary_table, singles_table])
    else:
        binary_table = InitialBinaryTable.InitialBinaries(
            mass1_binary,
            mass2_binary,
            porb,
            ecc,
            tphysf,
            kstar1,
            kstar2,
            metallicity,
            binfrac=binfrac,
        )

    return (
        binary_table,
        m_sampled_singles,
        m_sampled_binaries,
        n_singles,
        n_binaries,
    )


register_sampler(
    "independent",
    InitialBinaryTable,
    get_independent_sampler,
    usage="final_kstar1, final_kstar2, binfrac_model, primary_model, ecc_model, SFH_model, component_age, metallicity, size",
)


class Sample(object):
    # sample primary masses
    def sample_primary(self, primary_model='kroupa01', size=None, **kwargs):
        """Sample the primary mass (always the most massive star) from a user-selected model

            kroupa93 follows Kroupa (1993), normalization comes from
            `Hurley 2002 <https://arxiv.org/abs/astro-ph/0201220>`_
            between 0.08 and 150 Msun
            salpter55 follows
            `Salpeter (1955) <http://adsabs.harvard.edu/abs/1955ApJ...121..161S>`_
            between 0.08 and 150 Msun
            kroupa01 follows Kroupa (2001) <https://arxiv.org/abs/astro-ph/0009005>
            between 0.08 and 100 Msun


            Parameters
            ----------
            primary_model : str, optional
            model for mass distribution; choose from:

            kroupa93 follows Kroupa (1993), normalization comes from
            `Hurley 2002 <https://arxiv.org/abs/astro-ph/0201220>`_
            valid for masses between 0.1 and 100 Msun

            salpter55 follows
            `Salpeter (1955) <http://adsabs.harvard.edu/abs/1955ApJ...121..161S>`_
            valid for masses between 0.1 and 100 Msun

            kroupa01 follows Kroupa (2001), normalization comes from
            `Hurley 2002 <https://arxiv.org/abs/astro-ph/0009005>`_
            valid for masses between 0.1 and 100 Msun

            custom is a generic piecewise power law that takes in the power
            law slopes and break points given in the optional input lists (alphas, mcuts)
            default alphas and mcuts yield an IMF identical to kroupa01

            Default kroupa01

            size : int, optional
            number of initial primary masses to sample
            NOTE: this is set in cosmic-pop call as Nstep

            alphas : array, optional
            absolute values of the power law slopes for primary_model = 'custom'
            Default [-1.3,-2.3,-2.3] (identical to slopes for primary_model = 'kroupa01')

            mcuts : array, optional, units of Msun
            break points separating the power law 'pieces' for primary_model = 'custom'
            Default [0.08,0.5,1.0,150.] (identical to breaks for primary_model = 'kroupa01')

            Optional kwargs are defined in `get_independent_sampler`

            Returns
            -------
            a_0 : array
            Sampled primary masses
            np.sum(a_0) : float
            Total amount of mass sampled
            """

        # Read in m1_min and m_max kwargs, if provided
        m1_min = kwargs["m1_min"] if "m1_min" in kwargs.keys() else 0.08
        m_max = kwargs["m_max"] if "m_max" in kwargs.keys() else 150.0

        # Make sure m1_min value is below 0.5, since otherwise it will not work for Kroupa IMF
        if m1_min > 0.5:
            raise ValueError("m1_min must be greater than 0.5 Msun")

        if primary_model == 'kroupa93':
            alphas, mcuts = [-1.3,-2.2,-2.7], [m1_min,0.5,1.0,m_max]
        # Since COSMIC/BSE can't handle < 0.08Msun, we by default truncate at 0.08 Msun instead of 0.01
        elif primary_model == 'kroupa01':
            alphas, mcuts = [-1.3,-2.3], [m1_min,0.5,m_max]
        elif primary_model == 'salpeter55':
            alphas, mcuts = [-2.35], [m1_min,m_max]
        elif primary_model == 'custom':
            if 'alphas' in kwargs and 'mcuts' in kwargs:
                alphas = kwargs.pop("alphas", [-1.3,-2.3,-2.3])
                mcuts = kwargs.pop("mcuts", [m1_min,0.5,1.0,m_max])
            else:
                raise ValueError("You must supply both alphas and mcuts to use"
                                 " a custom IMF generator")

        Ncumulative, Ntotal, coeff = [], 0., 1.
        for i in range(len(alphas)):
            g = 1. + alphas[i]
            # Compute this piece of the IMF's contribution to Ntotal
            if alphas[i] == -1: Ntotal += coeff * np.log(mcuts[i+1]/mcuts[i])
            else: Ntotal += coeff/g * (mcuts[i+1]**g - mcuts[i]**g)
            Ncumulative.append(Ntotal)
            if i < len(alphas)-1: coeff *= mcuts[i+1]**(-alphas[i+1]+alphas[i])

        cutoffs = np.array(Ncumulative)/Ntotal
        u = np.random.uniform(0.,1.,size)
        idxs = [() for i in range(len(alphas))]

        for i in range(len(alphas)):
            if i == 0: idxs[i], = np.where(u <= cutoffs[0])
            elif i < len(alphas)-1: idxs[i], = np.where((u > cutoffs[i-1]) & (u <= cutoffs[i]))
            else: idxs[i], = np.where(u > cutoffs[i-1])
        for i in range(len(alphas)):
            if alphas[i] == -1.0:
                u[idxs[i]] = 10**np.random.uniform(np.log10(mcuts[i]),
                                                   np.log10(mcuts[i+1]),
                                                   len(idxs[i]))
            else:
                u[idxs[i]] = utils.rndm(a=mcuts[i], b=mcuts[i+1], g=alphas[i], size=len(idxs[i]))

        return u, np.sum(u)

    # sample secondary mass
    def sample_secondary(self, primary_mass, **kwargs):
        """Sample a secondary mass using draws from a uniform mass ratio distribution motivated by
        `Mazeh et al. (1992) <http://adsabs.harvard.edu/abs/1992ApJ...401..265M>`_
        and `Goldberg & Mazeh (1994) <http://adsabs.harvard.edu/abs/1994ApJ...429..362G>`_

        NOTE: the lower lim is set by either qmin or m2_min which are passed as kwargs

        Parameters
        ----------
        primary_mass : `array`
            sets the maximum secondary mass (for a maximum mass ratio of 1)

        Optional kwargs are defined in `get_independent_sampler`

        Returns
        -------
        secondary_mass : array
            sampled secondary masses with array size matching size of
            primary_mass
        """

        qmin = kwargs["qmin"] if "qmin" in kwargs.keys() else 0.0
        m1_min = kwargs["m1_min"] if "m1_min" in kwargs.keys() else 0.08
        m2_min = kwargs["m2_min"] if "m2_min" in kwargs.keys() else None
        if (m2_min is None) & (qmin is None):
            warnings.warn("It is highly recommended that you specify either qmin or m2_min!")
        if (m2_min is not None) and (m2_min > m1_min):
            raise ValueError("The m2_min you specified is above the minimum"
                             " primary mass of the IMF, either lower m2_min or"
                             " raise the lower value of your sampled primaries")

        # --- `msort` kwarg can be set to have different qmin above `msort`
        msort = kwargs["msort"] if "msort" in kwargs.keys() else None
        qmin_msort = kwargs["qmin_msort"] if "qmin_msort" in kwargs.keys() else None
        m2_min_msort = kwargs["m2_min_msort"] if "m2_min_msort" in kwargs.keys() else None
        if (msort is None) and (qmin_msort is not None):
            raise ValueError("If qmin_msort is specified, you must also supply a value for msort")
        if (msort is None) and (m2_min_msort is not None):
            raise ValueError("If m2_min_msort is specified, you must also supply a value for msort")
        if (m2_min_msort is not None) and (m2_min_msort > msort):
            raise ValueError("The m2_min_msort you specified is above the minimum"
                             " primary mass of the high-mass binaries msort")

        if (msort is not None) and (qmin_msort is not None):
            (highmassIdx,) = np.where(primary_mass >= msort)
            (lowmassIdx,) = np.where(primary_mass < msort)
        else:
            (highmassIdx,) = np.where(primary_mass < 0)
            (lowmassIdx,) = np.where(primary_mass >= 0)   # all idxs


        qmin_vals = -1 * np.ones_like(primary_mass)
        # --- qmin for low-mass systems (all systems if msort is not specified)
        if (qmin > 0.0):
            qmin_vals[lowmassIdx] = qmin * np.ones_like(primary_mass[lowmassIdx])
        elif (qmin < 0.0):
            # mass-dependent qmin, assume qmin=0.1 for m_primary<5
            dat = np.array([[5.0, 0.1363522012578616],
                            [6.999999999999993, 0.1363522012578616],
                            [12.599999999999994, 0.11874213836477984],
                            [20.999999999999993, 0.09962264150943395],
                            [29.39999999999999, 0.0820125786163522],
                            [41, 0.06490566037735851],
                            [55, 0.052327044025157254],
                            [70.19999999999999, 0.04301886792452836],
                            [87.4, 0.03622641509433966],
                            [107.40000000000002, 0.030188679245283068],
                            [133.40000000000003, 0.02515723270440262],
                            [156.60000000000002, 0.02163522012578628],
                            [175.40000000000003, 0.01962264150943399],
                            [200.20000000000005, 0.017358490566037776]])
            from scipy.interpolate import interp1d
            qmin_interp = interp1d(dat[:, 0], dat[:, 1])
            qmin_vals[lowmassIdx] = np.ones_like(primary_mass[lowmassIdx]) * 0.1
            ind_5, = np.where(primary_mass[lowmassIdx] > 5.0)
            qmin_vals[lowmassIdx][ind_5] = qmin_interp(primary_mass[lowmassIdx][ind_5])
        else:
            qmin_vals[lowmassIdx] = np.zeros_like(primary_mass[lowmassIdx])
        # --- qmin for high-mass systems, if msort and qmin_msort are specified
        if (msort is not None) and (qmin_msort is not None):
            if (qmin_msort > 0.0):
                qmin_vals[highmassIdx] = qmin_msort * np.ones_like(primary_mass[highmassIdx])
            elif (qmin_msort < 0.0):
                # mass-dependent qmin, assume qmin=0.1 for m_primary<5
                dat = np.array([[5.0, 0.1363522012578616],
                                [6.999999999999993, 0.1363522012578616],
                                [12.599999999999994, 0.11874213836477984],
                                [20.999999999999993, 0.09962264150943395],
                                [29.39999999999999, 0.0820125786163522],
                                [41, 0.06490566037735851],
                                [55, 0.052327044025157254],
                                [70.19999999999999, 0.04301886792452836],
                                [87.4, 0.03622641509433966],
                                [107.40000000000002, 0.030188679245283068],
                                [133.40000000000003, 0.02515723270440262],
                                [156.60000000000002, 0.02163522012578628],
                                [175.40000000000003, 0.01962264150943399],
                                [200.20000000000005, 0.017358490566037776]])
                from scipy.interpolate import interp1d
                qmin_interp = interp1d(dat[:, 0], dat[:, 1])
                qmin_vals[highmassIdx] = np.ones_like(primary_mass[highmassIdx]) * 0.1
                ind_5, = np.where(primary_mass[highmassIdx] > 5.0)
                qmin_vals[highmassIdx][ind_5] = qmin_interp(primary_mass[highmassIdx][ind_5])
            else:
                qmin_vals[highmassIdx] = np.zeros_like(primary_mass[highmassIdx])

        # --- apply m2_min and m2_min_msort, if specified
        if m2_min is not None:
            qmin_vals[lowmassIdx] = np.maximum(qmin_vals[lowmassIdx], m2_min/primary_mass[lowmassIdx])
        if m2_min_msort is not None:
            qmin_vals[highmassIdx] = np.maximum(qmin_vals[highmassIdx], m2_min_msort/primary_mass[highmassIdx])

        # --- now, randomly sample mass ratios and get secondary masses
        secondary_mass = np.random.uniform(qmin_vals, 1) * primary_mass



        return secondary_mass

    def binary_select(self, primary_mass, binfrac_model=0.5, **kwargs):
        """Select which primary masses will have a companion using
        either a binary fraction specified by a float or a
        primary-mass dependent binary fraction following
        `van Haaften et al.(2009) <http://adsabs.harvard.edu/abs/2013A%26A...552A..69V>`_ in appdx

        Parameters
        ----------
            primary_mass : array
                Mass that determines the binary fraction

            binfrac_model : str or float
                vanHaaften - primary mass dependent and ONLY VALID up to 100 Msun
                float - fraction of binaries; 0.5 means 2 in 3 stars are a binary pair while 1
                means every star is in a binary pair

        Optional kwargs are defined in `get_independent_sampler`

        Returns
        -------
            stars_in_binary : array
                primary masses that will have a binary companion

            stars_in_single : array
                primary masses that will be single stars

            binary_fraction : array
                system-specific probability of being in a binary

            binaryIdx : array
                Idx of stars in binary
        """

        # --- `msort` kwarg can be set to have different binary fraction above `msort`
        msort = kwargs["msort"] if "msort" in kwargs.keys() else None
        binfrac_model_msort = kwargs["binfrac_model_msort"] if "binfrac_model_msort" in kwargs.keys() else None
        if (msort is None) and (binfrac_model_msort is not None):
            raise ValueError("If binfrac_model_msort is specified, you must also supply a value for msort")
        if (msort is not None) and (binfrac_model_msort is not None):
            (highmassIdx,) = np.where(primary_mass >= msort)
            (lowmassIdx,) = np.where(primary_mass < msort)
        else:
            (highmassIdx,) = np.where(primary_mass < 0)
            (lowmassIdx,) = np.where(primary_mass >= 0)   # all idxs


        # --- read in binfrac models
        if type(binfrac_model) == str:
            if binfrac_model == "vanHaaften":
                binary_fraction_low = 1 / 2.0 + 1 / \
                    4.0 * np.log10(primary_mass[lowmassIdx])
                binary_choose_low = np.random.uniform(
                    0, 1.0, primary_mass[lowmassIdx].size)

                (singleIdx_low,) = np.where(
                    binary_fraction_low < binary_choose_low)
                (binaryIdx_low,) = np.where(
                    binary_fraction_low >= binary_choose_low)
            else:
                raise ValueError(
                    "You have supplied a non-supported binary fraction model. Please choose vanHaaften or a float"
                )
        elif type(binfrac_model) == float:
            if (binfrac_model <= 1.0) & (binfrac_model >= 0.0):
                binary_fraction_low = binfrac_model * \
                    np.ones(primary_mass[lowmassIdx].size)
                binary_choose_low = np.random.uniform(
                    0, 1.0, primary_mass[lowmassIdx].size)

                (singleIdx_low,) = np.where(
                    binary_choose_low > binary_fraction_low)
                (binaryIdx_low,) = np.where(
                    binary_choose_low <= binary_fraction_low)
            else:
                raise ValueError(
                    "You have supplied a fraction outside of 0-1. Please choose a fraction between 0 and 1."
                )
        else:
            raise ValueError(
                "You have not supplied a model or a fraction. Please choose either vanHaaften or a float"
            )

        # --- if using a different binary fraction for high-mass systems
        if (binfrac_model_msort is not None) and (type(binfrac_model_msort) == str):
            if binfrac_model_msort == "vanHaaften":
                binary_fraction_high = 1 / 2.0 + 1 / \
                    4.0 * np.log10(primary_mass[highmassIdx])
                binary_choose_high = np.random.uniform(
                    0, 1.0, primary_mass[highmassIdx].size)

                (singleIdx_high,) = np.where(
                    binary_fraction_high < binary_choose_high)
                (binaryIdx_high,) = np.where(
                    binary_fraction_high >= binary_choose_high)
            else:
                raise ValueError(
                    "You have supplied a non-supported binary fraction model. Please choose vanHaaften or a float"
                )
        elif (binfrac_model_msort is not None) and (type(binfrac_model_msort) == float):
            if (binfrac_model_msort <= 1.0) & (binfrac_model_msort >= 0.0):
                binary_fraction_high = binfrac_model_msort * \
                    np.ones(primary_mass[highmassIdx].size)
                binary_choose_high = np.random.uniform(
                    0, 1.0, primary_mass[highmassIdx].size)

                (singleIdx_high,) = np.where(
                    binary_choose_high > binary_fraction_high)
                (binaryIdx_high,) = np.where(
                    binary_choose_high <= binary_fraction_high)
            else:
                raise ValueError(
                    "You have supplied a fraction outside of 0-1. Please choose a fraction between 0 and 1."
                )
        elif (binfrac_model_msort is not None):
            raise ValueError(
                "You have not supplied a model or a fraction. Please choose either vanHaaften or a float"
            )


        # --- get pertinent info
        if (binfrac_model_msort is not None):
            stars_in_binary = np.append(
                primary_mass[highmassIdx][binaryIdx_high], primary_mass[lowmassIdx][binaryIdx_low])
            stars_in_single = np.append(
                primary_mass[highmassIdx][singleIdx_high], primary_mass[lowmassIdx][singleIdx_low])
            binary_fraction = np.append(
                binary_fraction_high[binaryIdx_high], binary_fraction_low[binaryIdx_low])
            binaryIdx = np.append(
                highmassIdx[binaryIdx_high], lowmassIdx[binaryIdx_low])
        else:
            stars_in_binary = primary_mass[lowmassIdx][binaryIdx_low]
            stars_in_single = primary_mass[lowmassIdx][singleIdx_low]
            binary_fraction = binary_fraction_low[binaryIdx_low]
            binaryIdx = lowmassIdx[binaryIdx_low]

        return (
            stars_in_binary,
            stars_in_single,
            binary_fraction,
            binaryIdx,
        )

    def sample_porb(self, mass1, mass2, rad1, rad2, porb_model, porb_max=None, size=None, **kwargs):
        """Sample the orbital period according to the user-specified model

        Parameters
        ----------
        mass1 : array
            primary masses
        mass2 : array
            secondary masses
        rad1 : array
            radii of the primaries. 
        rad2 : array
            radii of the secondaries 
        porb_model : string
            selects which model to sample orbital periods, choices include:
            log_uniform : semi-major axis flat in log space from RRLO < 0.5 up to 1e5 Rsun according to
            `Abt (1983) <http://adsabs.harvard.edu/abs/1983ARA%26A..21..343A>`_
            and consistent with Dominik+2012,2013
            and then converted to orbital period in days using Kepler III
            sana12 : power law orbital period between 0.15 < log(P/day) < 5.5 following
            `Sana+2012 <https://ui.adsabs.harvard.edu/abs/2012Sci...337..444S/abstract>_`
            renzo19 : power law orbital period for m1 > 15Msun binaries from
            `Sana+2012 <https://ui.adsabs.harvard.edu/abs/2012Sci...337..444S/abstract>_`
            following the implementation of
            `Renzo+2019 <https://ui.adsabs.harvard.edu/abs/2019A%26A...624A..66R/abstract>_`
            and flat in log otherwise
            raghavan10 : log normal orbital periods in days with mean_logP = 4.9 
            and sigma_logP = 2.3 between 0 < log10(P/day) < 9  following 
            `Raghavan+2010 <https://ui.adsabs.harvard.edu/abs/2010ApJS..190....1R/abstract>_`
            moe19 : log normal orbital periods in days with mean_logP = 4.9 
            and sigma_logP = 2.3 between 0 < log10(P/day) < 9  following 
            `Raghavan+2010 <https://ui.adsabs.harvard.edu/abs/2010ApJS..190....1R/abstract>_`
            but with different close binary fractions following 
            `Moe+2019 <https://ui.adsabs.harvard.edu/abs/2019ApJ...875...61M/abstract>_`
        met : float
            metallicity of the population

        Returns
        -------
        porb : array
            orbital period with array size equalling array size
            of mass1 and mass2 in units of days
        aRL_over_a: array
            ratio of radius where RL overflow starts to the sampled seperation
            used to truncate the eccentricitiy distribution
        """

        # First we need to compute where RL overflow starts.  We truncate the lower-bound
        # of the period distribution there
        q = mass2 / mass1
        RL_fac = (0.49 * q ** (2.0 / 3.0)) / (
            0.6 * q ** (2.0 / 3.0) + np.log(1 + q ** 1.0 / 3.0)
        )

        q2 = mass1 / mass2
        RL_fac2 = (0.49 * q2 ** (2.0 / 3.0)) / (
            0.6 * q2 ** (2.0 / 3.0) + np.log(1 + q2 ** 1.0 / 3.0)
        )

        # include the factor for the eccentricity
        RL_max = 2 * rad1 / RL_fac
        (ind_switch,) = np.where(RL_max < 2 * rad2 / RL_fac2)
        if len(ind_switch) >= 1:
            RL_max[ind_switch] = 2 * rad2[ind_switch] / RL_fac2[ind_switch]

        # Can either sample the porb first and truncate the eccentricities at RL overflow
        # or sample the eccentricities first and truncate a(1-e) at RL overflow
        #
        # If we haven't sampled the eccentricities, then the minimum semi-major axis is at
        # RL overflow
        #
        # If we have, then the minimum pericenter is set to RL overflow
        a_min = RL_max 

        if porb_model == "log_uniform":
            if porb_max is None:
                a_0 = np.random.uniform(np.log(a_min), np.log(1e5), size)
            else:
                # If in CMC, only sample binaries as wide as the local hard/soft boundary
                a_max = utils.a_from_p(porb_max,mass1,mass2) 
                a_max[a_max < a_min] = a_min[a_max < a_min]
                a_0 = np.random.uniform(np.log(a_min), np.log(a_max), size)

            # convert out of log space
            a_0 = np.exp(a_0)
            aRL_over_a = a_min/a_0

            # convert to au
            rsun_au = 0.00465047
            a_0 = a_0 * rsun_au

            # convert to orbital period in years
            yr_day = 365.24
            porb_yr = ((a_0 ** 3.0) / (mass1 + mass2)) ** 0.5
            porb = porb_yr * yr_day
        elif porb_model == "sana12":
            # Same here: if using CMC, set the maximum porb to the smaller of either the
            # hard/soft boundary or 5.5 (from Sana paper)
            if porb_max is None:
                log10_porb_max = 5.5
            else:
                log10_porb_max = np.minimum(5.5,np.log10(porb_max))

            # Use the lower limit from the Sana12 distribution, unless this means the binaries are sampled at RL overflow. If so, 
            # change the lower limit to a_min
                         
            log10_porb_min = np.array([0.15]*len(a_min)) 
            RL_porb = utils.p_from_a(a_min,mass1,mass2)
            log10_RL_porb = np.log10(RL_porb)
            log10_porb_min[log10_porb_min <  log10_RL_porb] = log10_RL_porb[log10_porb_min < log10_RL_porb]
            
            porb = 10 ** utils.rndm(a=log10_porb_min, b=log10_porb_max, g=-0.55, size=size)
            aRL_over_a = a_min / utils.a_from_p(porb,mass1,mass2) 

        elif porb_model == "renzo19":
            # Same here: if using CMC, set the maximum porb to the smaller of either the
            # hard/soft boundary or 5.5 (from Sana paper)
            if porb_max is None:
                log10_porb_max = 5.5
            
            else:
                log10_porb_max = np.minimum(5.5,np.log10(porb_max))
            
            # Use the lower limit from the Sana12 distribution, unless this means the binaries are sampled at RL overflow. If so, 
            # change the lower limit to a_min
            log10_porb_min = np.array([0.15]*len(a_min)) 
            RL_porb = utils.p_from_a(a_min,mass1,mass2)
            log10_RL_porb = np.log10(RL_porb)
            log10_porb_min[log10_porb_min <  log10_RL_porb] = log10_RL_porb[log10_porb_min < log10_RL_porb]
            
            porb = 10 ** (np.random.uniform(log10_porb_min, log10_porb_max, size))
            (ind_massive,) = np.where(mass1 > 15)
            
            if type(log10_porb_max) != float:
                log10_porb_max = log10_porb_max[ind_massive]
                log10_porb_min = log10_porb_min[ind_massive]              
            
            
            porb[ind_massive] = 10 ** utils.rndm(
                a=log10_porb_min[ind_massive], b=log10_porb_max, g=-0.55, size=len(ind_massive))
            aRL_over_a = a_min / utils.a_from_p(porb,mass1,mass2) 
        
        elif porb_model == "raghavan10":
            import scipy
            # Same here: if using CMC, set the maximum porb to the smaller of either the
            # hard/soft boundary or 5.5 (from Sana paper)
            if porb_max is None:
                log10_porb_max = 9.0
            else:
                log10_porb_max = np.minimum(5.5, np.log10(porb_max))

            lower = 0
            upper = log10_porb_max
            mu = 4.9
            sigma = 2.3

            porb = 10 ** (scipy.stats.truncnorm.rvs(
                (lower-mu)/sigma,(upper-mu)/sigma, loc=mu, scale=sigma, size=size
                ))

            aRL_over_a = a_min / utils.a_from_p(porb,mass1,mass2)

        elif porb_model == "moe19":
            from scipy.interpolate import interp1d
            from scipy.stats import norm
            from scipy.integrate import trapz
            try:
                met = kwargs.pop('met')
            except:
                raise ValueError(
                    "You have chosen moe19 for the orbital period distribution which is a metallicity-dependent distribution. "
                    "Please specify a metallicity for the population."
                    )
            def get_logP_dist(nsamp, norm_wide, norm_close, mu=4.4, sigma=2.1):
                logP_lo_lim=0
                logP_hi_lim=9
                close_logP=4.0
                wide_logP=6.0
                neval = 500
                prob_wide = norm.pdf(np.linspace(wide_logP, logP_hi_lim, neval), loc=mu, scale=sigma)*norm_wide
                prob_close = norm.pdf(np.linspace(logP_lo_lim, close_logP, neval), loc=mu, scale=sigma)*norm_close
                slope = -(prob_close[-1] - prob_wide[0]) / (wide_logP - close_logP)
                prob_intermediate = slope * (np.linspace(close_logP, wide_logP, neval) - close_logP) + prob_close[-1]
                prob_interp_int = interp1d(np.linspace(close_logP, wide_logP, neval), prob_intermediate)

                log_p_success = []
                n_success = 0
                while n_success < nsamp:
                    logP_samp = np.random.uniform(logP_lo_lim, logP_hi_lim, nsamp*5)
                    logP_prob = np.random.uniform(0, 1, nsamp*5)
        
                    logP_samp_lo = logP_samp[logP_samp<close_logP]
                    logP_prob_lo = logP_prob[logP_samp<close_logP]
                    log_p_success.extend(logP_samp_lo[np.where(logP_prob_lo < norm.pdf(logP_samp_lo, loc=mu, scale=sigma)*norm_close)])
        
                    logP_samp_int = logP_samp[(logP_samp>=close_logP) & (logP_samp<wide_logP)]
                    logP_prob_int = logP_prob[(logP_samp>=close_logP) & (logP_samp<wide_logP)]
                    log_p_success.extend(logP_samp_int[np.where(logP_prob_int < prob_interp_int(logP_samp_int))])
    
                    logP_samp_hi = logP_samp[(logP_samp>=wide_logP)]
                    logP_prob_hi = logP_prob[(logP_samp>=wide_logP)]

                    log_p_success.extend(logP_samp_hi[np.where(logP_prob_hi < norm.pdf(logP_samp_hi, loc=mu, scale=sigma)*norm_wide)])

                    n_success = len(log_p_success)
                log_p_success = np.array(log_p_success)[np.random.randint(0,n_success,nsamp)]    
                return log_p_success
            norm_wide, norm_close = utils.get_porb_norm(met)
            logP_dist = get_logP_dist(size, norm_wide, norm_close)
            logP_dist = logP_dist[np.random.randint(0, len(logP_dist), size)]
            porb = 10**logP_dist 
            aRL_over_a = a_min / utils.a_from_p(porb,mass1,mass2) 
            

        else:
            raise ValueError(
                "You have supplied a non-supported model; Please choose either log_flat, sana12, renzo19, raghavan10, or moe19"
            )
        return porb, aRL_over_a 

    def sample_ecc(self, aRL_over_a, ecc_model="sana12", size=None):
        """Sample the eccentricity according to a user specified model

        Parameters
        ----------
        ecc_model : string
            'thermal' samples from a  thermal eccentricity distribution following
            `Heggie (1975) <http://adsabs.harvard.edu/abs/1975MNRAS.173..729H>`_
            'uniform' samples from a uniform eccentricity distribution
            'sana12' samples from the eccentricity distribution from
            `Sana+2012 <https://ui.adsabs.harvard.edu/abs/2012Sci...337..444S/abstract>_`
            'circular' assumes zero eccentricity for all systems
            DEFAULT = 'sana12'

        aRL_over_a : ratio of the minimum seperation (where RL overflow starts)
            to the sampled semi-major axis.  Use this to truncate the eccentricitiy

        size : int, optional
            number of eccentricities to sample
            this is set in cosmic-pop call as Nstep

        Returns
        -------
        ecc : array
            array of sampled eccentricities with size=size
        """

        # if we sampled the periods first, we need to truncate the eccentricities
        # to avoid RL overflow/collision at pericenter
        e_max = 1.0 - aRL_over_a

        if ecc_model == "thermal":
            a_0 = np.random.uniform(0.0, e_max**2, size)
            ecc = a_0 ** 0.5
            return ecc

        elif ecc_model == "uniform":
            ecc = np.random.uniform(0.0, e_max, size)
            return ecc

        elif ecc_model == "sana12":
            sana_max = np.array([0.9]*len(e_max))
            max_e = np.minimum(e_max, sana_max)
            ecc = utils.rndm(a=0.001, b=max_e, g=-0.45, size=size)
            
            return ecc

        elif ecc_model == "circular":
            ecc = np.zeros(size)
            return ecc

        else:
            raise ValueError("You have specified an unsupported model. Please choose from thermal, "
                             "uniform, sana12, or circular")

    def sample_SFH(self, SF_start=13700.0, SF_duration=0.0, met=0.02, size=None):
        """Sample an evolution time for each binary based on a user-specified
        time at the start of star formation and the duration of star formation.
        The default is a burst of star formation 13,700 Myr in the past.

        Parameters
        ----------
        SF_start : float
            Time in the past when star formation initiates in Myr
        SF_duration : float
            Duration of constant star formation beginning from SF_Start in Myr
        met : float
            metallicity of the population [Z_sun = 0.02]
            Default: 0.02
        size : int, optional
            number of evolution times to sample
            NOTE: this is set in cosmic-pop call as Nstep

        Returns
        -------
        tphys : array
            array of evolution times of size=size
        metallicity : array
            array of metallicities
        """

        if (SF_start > 0.0) & (SF_duration >= 0.0):
            tphys = np.random.uniform(SF_start - SF_duration, SF_start, size)
            metallicity = np.ones(size) * met
            return tphys, metallicity
        else:
            raise ValueError(
                'SF_start and SF_duration must be positive and SF_start must be greater than 0.0')

    def set_kstar(self, mass):
        """Initialize stellar types according to BSE classification
        kstar=1 if M>=0.7 Msun; kstar=0 if M<0.7 Msun

        Parameters
        ----------
        mass : array
            array of masses

        Returns
        -------
        kstar : array
            array of initial stellar types
        """

        kstar = np.zeros(mass.size)
        low_cutoff = 0.7
        lowIdx = np.where(mass < low_cutoff)[0]
        hiIdx = np.where(mass >= low_cutoff)[0]

        kstar[lowIdx] = 0
        kstar[hiIdx] = 1

        return kstar

    def set_reff(self, mass, metallicity, zsun=0.02):
        """
        Better way to set the radii from BSE, by calling it directly

        takes masses and metallicities, and returns the radii

        Note that the BSE function is hard-coded to go through arrays
        of length 10^5.  If your masses are more than that, you'll
        need to divide it into chunks
        """

        max_array_size = 100000
        total_length = len(mass)
        radii = np.zeros(total_length)

        _evolvebin.metvars.zsun = zsun

        idx = 0
        while total_length > max_array_size:
            ## cycle through the masses max_array_size number at a time
            temp_mass = mass[idx*max_array_size:(idx+1)*max_array_size]

            temp_radii = _evolvebin.compute_r(temp_mass,metallicity,max_array_size)

            ## put these in the radii array
            radii[idx*max_array_size:(idx+1)*max_array_size] = temp_radii

            total_length -= max_array_size
            idx += 1

        length_remaining = total_length

        ## if smaller than 10^5, need to pad out the array
        temp_mass = np.zeros(max_array_size)
        temp_mass[:length_remaining] = mass[-length_remaining:]

        temp_radii = _evolvebin.compute_r(temp_mass,metallicity,length_remaining)

        #finish up the array
        radii[-length_remaining:] = temp_radii[:length_remaining]

        return radii



# -*- coding: utf-8 -*-
# Copyright (C) Carl Rodriguez (2020 - 2021)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`Elson`
"""

from scipy.interpolate import interp1d, CubicSpline
from scipy.integrate import quad
from scipy.special import hyp2f1
from scipy.optimize import brentq
import numpy as np
from numpy.random import uniform, normal
from scipy.stats import maxwell

__author__ = "Carl Rodriguez <carllouisrodriguez@gmail.com>"
__credits__ = "Carl Rodriguez <carllouisrodriguez@gmail.com>"


def M_enclosed(r, gamma, rho_0):
    """
    Compute the mass enclosed in an Elson profile at radius r with slope
    gamma, central concentration rho_0, and assumed scale factor a = 1

    In practice, this is only used to sample the positions of stars, so rho_0 is
    just picked to normalize the distribution (i.e. rho_0 s.t. M_enclosed(rmax) = 1)
    """

    prefactor = 4 * np.pi * rho_0 * r * r * r / 3.0

    hypergeometric = hyp2f1(1.5, (gamma + 1.0) / 2.0, 2.5, -(r ** 2))

    return prefactor * hypergeometric

def phi_r(r, gamma, rho_0):
    """
    Compute the gravitational potential of an Elson profile at radius r with slope
    gamma, central concentration rho_0, and assumed scale factor a = 1

    Needed to compute the escape speed (for resampling stars)
    """

    prefactor = - 4 * np.pi * rho_0 / (gamma - 1)

    hypergeometric = hyp2f1(0.5, (gamma - 1.0) / 2.0, 1.5, -(r ** 2))

    return prefactor * hypergeometric

def rho_r(r, gamma, rho_0):
    """
    Compute the density of the Elson profile at radius r
    Best to use the same normalized rho_0 from M_enclosed
    """
    

    return rho_0 * pow(1 + r * r, -(gamma + 1.0) / 2.0)

def virial_radius_analytic(gamma, r_max):
    """
    Virial radius is best calculated directly, since rmax may be pretty far from
    infinity.  Directly integrate 4*pi*r*rho*m_enclosed from 0 to rMax to get the
    binding energy, then just divide 0.5 by that.
    """

    rho_0 = 1.0 / M_enclosed(r_max, gamma, 1)

    integral, error = quad(lambda r: rho_r(r, gamma, rho_0) * r * 4 * np.pi * M_enclosed(r, gamma, rho_0), 0, r_max)

    return 1 / 2.0 / integral


def find_rmax_vir(r_max, gamma):
    """
    This is a little tricky: because the virial radius of the Elson profile
    depends on the maximum radius, if the profile is very flat (e.g. gamma~2)
    then you need a large maximum radius to get a large number of virial radius
    in there.  This basically finds r such that  r / rVir(r) = r_max.  In other
    word, how far out do we need to go to have r_max number of virial radii in the
    sample.
    """
    rOrvirMin = 1.0
    rOrvirMax = 20000.0

    # rvir ~ 1 for gamma > 4, and the M_enclosed integral doesn't converge.
    # Here's a cheap fix for that...
    if gamma > 4:
        rOrvirMax /= 10


    def y_zero(r):
        return r / virial_radius_analytic(gamma, r) - r_max

    yMin = y_zero(rOrvirMin)
    yMax = y_zero(rOrvirMax)

    rmax_vir = brentq(y_zero, rOrvirMin,rOrvirMax)

    return rmax_vir


def find_sigma_sqr(r, r_max_cluster, gamma):
    """
    Find the 1D velocity dispersion at a given radius r using one of the
    spherial Jeans equations (and assuming velocity is isotropic)
    """

    rho_0 = 1.0 / M_enclosed(r_max_cluster, gamma, 1)

    jeans_integrand = (
        lambda rr: rho_r(rr, gamma, rho_0) * M_enclosed(rr, gamma, rho_0) / rr / rr
    )

    integral, error = quad(jeans_integrand, r, r_max_cluster)

    return integral / rho_r(r, gamma, rho_0)


def get_positions(N, r_max_cluster, gamma):
    """
    This one's easy: the mass enclosed function is just the CDF of the mass
    density, so just invert that, and you've got positions.
    """

    # First normalize the central density s.t. M_enc(r_max) = 1
    rho_0 = 1.0 / M_enclosed(r_max_cluster, gamma, 1)

    radii_grid = np.logspace(-3, np.log10(r_max_cluster), 1000)

    # Add r=0 to the array
    radii_grid[0] = 0

    mass_enclosed_grid = M_enclosed(radii_grid, gamma, rho_0)

    # if it's a steep profile, you'll get a lot of similar numbers (to machine precision)
    # remove them here
    mass_enclosed_grid, unique_idx = np.unique(mass_enclosed_grid,return_index=True)
    radii_grid = radii_grid[unique_idx]

    # Use an interpolator, but flip x and y for the CDF
    interpolator = interp1d(mass_enclosed_grid, radii_grid, kind="cubic")

    X = uniform(size=N)

    positions = interpolator(X)

    return positions

def get_velocities_old(r, r_max_cluster, gamma):
    """
    Uses the spherical Jeans functions to sample the velocity dispersion for the
    cluster at different radii, then draws a random, isotropic velocity for each
    star

    NOTE: this gives you correct velocity dispersion and produces a cluster in virial
    equilibrium, but it's not strictly speaking correct (the distribution should
    have some kurtosis, in addition to getting the variance of the Gaussian correct).
    You can see the disagreement in the tail of the distributions when sampling a
    Plummer sphere.  This is what mcluster does...

    Use the new get_velocities function, which generates a distribution function
    directly from rho and samples velocities from it

    This is kept around because it's super useful to sample from when the rejection sampling
    gets too slow at the last X samples (e.g. gamma ~ 10)

    returns (vr,vt) with same length as r
    """

    N = len(r)

    rho_0 = 1.0 / M_enclosed(r_max_cluster, gamma, 1)

    # Rather than calculate the integral for every stellar position, just
    # create an interpolator of 1000 or so points, then sample from the
    # interpolated curve
    radii_grid = np.logspace(np.log10(min(r) * 0.999), np.log10(max(r) * 1.001), 1000)
    sigma_sqr_grid = np.array(
        [find_sigma_sqr(rr, r_max_cluster, gamma) for rr in radii_grid]
    )

    interpolator = interp1d(radii_grid, sigma_sqr_grid, kind="cubic")

    # Draw the 1D velocity dispersions
    sigma = np.sqrt(interpolator(r))

    # Then sample an isotropic gaussian in vx,vy,vz using those dispersions
    vx = normal(scale=sigma, size=N)
    vy = normal(scale=sigma, size=N)
    vz = normal(scale=sigma, size=N)
    
    # Finally the rejection sampling: because of the Gaussian, some of the velocities
    # we draw will naturally be above the local escape speed of the cluster.  To avoid 
    # this, we flag them, and resample them from the velocity distribution
    while False:
        ve = np.sqrt(-2*phi_r(r, gamma, rho_0))
        escapers = (np.sqrt(vx**2 + vy**2 + vz**2) > ve)

        number_of_escapers = np.sum(escapers)
        if number_of_escapers == 0:
            break
        else:
            vx_temp = normal(scale=sigma[escapers], size=number_of_escapers)
            vy_temp = normal(scale=sigma[escapers], size=number_of_escapers)
            vz_temp = normal(scale=sigma[escapers], size=number_of_escapers)

        vx[escapers] = vx_temp
        vy[escapers] = vy_temp
        vz[escapers] = vz_temp

    return np.sqrt(vx**2 + vy**2 + vz**2) 

def get_velocities(r, r_max_cluster, gamma):
    """
    The correct way to generate velocities: generate the distribution function from rho,
    then use rejection sampling.

    returns (vr,vt) with same length as r
    """

    N = len(r)

    # If we actually want a plummer sphere, we can default to the simpler function
    # using the analytic distribution function
    if gamma == 4:
        return get_velocities_plummer(r,r_max_cluster)
    # Otherwise need to compute f(E) numerically

    # First get phi and rho over 100 points or so
    rho_0 = 1.0 / M_enclosed(max(r), gamma, 1)

    r_stensil = np.logspace(np.log10(min(r)),np.log10(max(r)),100)
    phi = phi_r(r_stensil,gamma,rho_0)
    rho = rho_r(r_stensil,gamma,rho_0)

    # Then construct a cubic spline of rho as a function of phi
    # Technically should be psi = -phi, but FITPACK only accepts increasing values
    rho_phi_spline = CubicSpline(phi,rho,extrapolate=False)

    f_E = []
    energies = []

    # Now integrate over energies to get the distribution function 
    for en in np.linspace(min(-phi),max(-phi),100):

        integrand = lambda psi: rho_phi_spline(-psi,2) / np.sqrt(en - psi)

        # Eqn 4.46b of Binney and Tremain (2nd Edition)
        # (4.140b in 1st edition)
        temp_f_E = ((quad(integrand,min(-phi)*1.01,en,limit=200)[0] +
                     rho_phi_spline(-min(-phi),1)/np.sqrt(en)) / 17.493418) # sqrt(8)*pi^2

        # Extrapolating derivatives from a cubic spline is hella dangerous
        # Better to have it throw NaNs and discard them
        if np.isnan(temp_f_E):
            continue
        else:
            f_E.append(temp_f_E)
            energies.append(en)

    # Set f(E) to zero at zero energy 
    f_E = np.array([0] + f_E)
    energies = np.array([0] + energies)

    # Make an interpolator for f(E) and compute f and phi at every star
    f_E_interp = interp1d(energies,f_E)
    phi_at_r = phi_r(r,gamma,rho_0)
    f_E_at_r = f_E_interp(-phi_at_r)

    # Make an initial guess for every velocity
    x_rand = uniform(size=N)
    y_rand = uniform(size=N)
    velocities = x_rand*np.sqrt(-2*phi_at_r)

    resample = np.ones(N,dtype=bool)
    number_to_resample = N

    # Now do the rejection sampling
    while number_to_resample > 0: 
        resample[resample] = (y_rand[resample]*f_E_at_r[resample] > 
                              x_rand[resample]**2 * f_E_interp(-phi_at_r[resample]*(1 - x_rand[resample]**2)))
        number_to_resample = np.sum(resample)
        x_rand[resample] = uniform(size=number_to_resample)
        y_rand[resample] = uniform(size=number_to_resample)

    velocities = x_rand * np.sqrt(-2*phi_at_r)

    theta = np.arccos(uniform(-1,1,N))

    vr = velocities*np.cos(theta)
    vt = velocities*np.sin(theta)

    return vr, vt

def get_velocities_plummer(r, r_max_cluster):
    """
    The correct way to generate velocities

    this is a special case for gamma=4, which is the plummer sphere (and has an analytic distribution function)

    returns (vr,vt) with same length as r
    """

    N = len(r)


    # Make an initial guess for every velocity
    x_rand = uniform(size=N)
    y_rand = uniform(size=N)
    vesc = np.sqrt(2/np.sqrt(1+r**2))
    velocities = x_rand * vesc

    resample = np.ones(N,dtype=bool)
    number_to_resample = N

    # Now do the rejection sampling; this is taken from Aarseth's original algorithm using the dist function
    while number_to_resample > 0: 
        resample[resample] = 0.1*y_rand[resample] > x_rand[resample]**2 * (1 - x_rand[resample]**2)**3.5
        number_to_resample = np.sum(resample)
        x_rand[resample] = uniform(size=number_to_resample)
        y_rand[resample] = uniform(size=number_to_resample)

    # Scale up to the local escape speed
    velocities = x_rand * vesc 

    theta = np.arccos(uniform(-1,1,N))

    vr = velocities*np.cos(theta)
    vt = velocities*np.sin(theta)

    return vr, vt


def scale_pos_and_vel(r,vr,vt):
    """
    Scale the positions and velocities to be in N-body units
    If we add binaries we'll do this again in initialcmctable.py

    takes r, vr, and vt as input

    returns (r,vr,vt) scaled to Henon units
    """

    # Normalize masses to 1/Mtotal
    mass = np.ones_like(r)/len(r)
    cumul_mass = np.cumsum(mass)

    radius = np.array(r)
    radius_p1 = np.append(radius[1:], [1e100])

    # Then compute the total kinetic and potential energy
    # There's probably a cleaner way to do the PE (this is a one-line version
    #  of the for loop we use in CMC; vectorized and pythonic, but sloppy)
    KE = 0.5 * sum(mass * (vr ** 2 + vt ** 2))
    PE = 0.5 * sum(
        mass[::-1]
        * np.cumsum((cumul_mass * (1.0 / radius - 1.0 / radius_p1))[::-1])
    )

    # Compute the position and velocity scalings
    rfac = 2 * PE
    vfac = 1.0 / np.sqrt(4 * KE)

    return r*rfac, vr*vfac, vt*vfac


def draw_r_vr_vt(N=100000, r_max=300, gamma=4):
    """
    Draw random velocities and positions from the Elson profile.

    N     = number of stars
    r_max = maximum number of virial radii for the farthest star
    gamma = steepness function of the profile

    Note that gamma=4 is the Plummer profile

    returns (vr,vt,r) in G=M_cluster=1 units

    N.B. if you're getting a "Expect x to not have duplicates" error
    with a large gamma, use a smaller r_max. 
    """
    # First convert r_max into max number of  virial radii
    r_max = find_rmax_vir(r_max, gamma)

    # Then draw the positions from the cumulative mass function
    r = get_positions(N, r_max, gamma)

    # Sort the radii (needed later)
    r = np.sort(r)

    # Finally, draw the velocities using the radii and one of the Jeans
    # equations
    vr, vt = get_velocities(r, r_max, gamma)

    # And scale them to Henon units 
    r,vr,vt = scale_pos_and_vel(r,vr,vt)

    return r, vr, vt



"""Unit test for cosmic
"""

__author__ = 'Katie Breivik <katie.breivik@gmail.com>'

import os
import unittest
import numpy as np
import pandas as pd

from ..evolve import Evolve
from ..sample.initialbinarytable import InitialBinaryTable
from ..sample import initialbinarytable
from .. import evolve

import warnings
warnings.filterwarnings("ignore")

TEST_DATA_DIR = os.path.join(os.path.split(__file__)[0], 'data')
PARAMS_INI = os.path.join(TEST_DATA_DIR,'Params.ini')
INIT_CONDITIONS = pd.read_hdf(os.path.join(TEST_DATA_DIR, 'initial_conditions_for_testing.hdf5'), key='initC')

init_conds_columns = initialbinarytable.INITIAL_CONDITIONS_COLUMNS_ALL

INIT_CONDITIONS_NO_BSE_COLUMNS = INIT_CONDITIONS[init_conds_columns]
BPP_DF = pd.read_hdf(os.path.join(TEST_DATA_DIR, 'unit_tests_results.hdf5'), key='bpp')
BCM_DF = pd.read_hdf(os.path.join(TEST_DATA_DIR, 'unit_tests_results.hdf5'), key='bcm')
BSEFlag_columns = list(set(evolve.INITIAL_BINARY_TABLE_SAVE_COLUMNS) - set(initialbinarytable.INITIAL_CONDITIONS_COLUMNS_ALL)) 
BSEDict = INIT_CONDITIONS[BSEFlag_columns].to_dict(orient='index')[0]
BSEDict['qcrit_array'] = [0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0]
BSEDict['natal_kick_array'] = [[-100.0,-100.0,-100.0,-100.0,0.0], [-100.0,-100.0,-100.0,-100.0,0.0]]
BSEDict['fprimc_array'] = [2.0/21.0,2.0/21.0,2.0/21.0,2.0/21.0,2.0/21.0,
                           2.0/21.0,2.0/21.0,2.0/21.0,2.0/21.0,2.0/21.0,
                           2.0/21.0,2.0/21.0,2.0/21.0,2.0/21.0,2.0/21.0,2.0/21.0]
BSEDict['grflag'] = 1
BSEDict['don_lim'] = -1
BSEDict['acc_lim'] = -1


class TestEvolve(unittest.TestCase):
    """`TestCase` for the cosmic
    """
    def test_single_evolve_with_table(self):

        # Check that the sample_primary function samples mass correctly
        EvolvedBinaryBPP, EvolvedBinaryBCM, initCond, kick_info = Evolve.evolve(
            initialbinarytable=INIT_CONDITIONS, randomseed=523574)

        pd.testing.assert_frame_equal(EvolvedBinaryBPP, BPP_DF, check_dtype=False, check_exact=False)
        pd.testing.assert_frame_equal(EvolvedBinaryBCM, BCM_DF, check_dtype=False, check_exact=False)

    def test_single_evolve_with_dict(self):
        EvolvedBinaryBPP, EvolvedBinaryBCM, initCond, kick_info = Evolve.evolve(
            initialbinarytable=INIT_CONDITIONS_NO_BSE_COLUMNS, BSEDict=BSEDict, randomseed=523574)

        pd.testing.assert_frame_equal(EvolvedBinaryBPP, BPP_DF, check_dtype=False, check_exact=False)
        pd.testing.assert_frame_equal(EvolvedBinaryBCM, BCM_DF, check_dtype=False, check_exact=False)

    def test_single_evolve_with_inifile(self):
        EvolvedBinaryBPP, EvolvedBinaryBCM, initCond, kick_info = Evolve.evolve(
            initialbinarytable=INIT_CONDITIONS_NO_BSE_COLUMNS, params=PARAMS_INI, randomseed=523574)

        pd.testing.assert_frame_equal(EvolvedBinaryBPP, BPP_DF, check_dtype=False, check_exact=False)
        pd.testing.assert_frame_equal(EvolvedBinaryBCM, BCM_DF, check_dtype=False, check_exact=False)

    def test_single_evolve_with_dict_and_table(self):
        EvolvedBinaryBPP, EvolvedBinaryBCM, initCond, kick_info = Evolve.evolve(
            initialbinarytable=INIT_CONDITIONS, BSEDict=BSEDict, randomseed=523574)

        pd.testing.assert_frame_equal(EvolvedBinaryBPP, BPP_DF, check_dtype=False, check_exact=False)
        pd.testing.assert_frame_equal(EvolvedBinaryBCM, BCM_DF, check_dtype=False, check_exact=False)

    def test_multi_evolve_with_table(self):
        # Check that the sample_primary function samples mass correctly
        EvolvedBinaryBPP, EvolvedBinaryBCM, initCond, kick_info = Evolve.evolve(
            initialbinarytable=INIT_CONDITIONS, n_per_block=100)

        pd.testing.assert_frame_equal(EvolvedBinaryBPP, BPP_DF, check_dtype=False, check_exact=False)
        pd.testing.assert_frame_equal(EvolvedBinaryBCM, BCM_DF, check_dtype=False, check_exact=False)

    def test_multi_evolve_with_dict(self):
        EvolvedBinaryBPP, EvolvedBinaryBCM, initCond, kick_info = Evolve.evolve(
            initialbinarytable=INIT_CONDITIONS_NO_BSE_COLUMNS, BSEDict=BSEDict, randomseed=523574, n_per_block=100)

        pd.testing.assert_frame_equal(EvolvedBinaryBPP, BPP_DF, check_dtype=False, check_exact=False)
        pd.testing.assert_frame_equal(EvolvedBinaryBCM, BCM_DF, check_dtype=False, check_exact=False)

    def test_multi_evolve_with_inifile(self):
        EvolvedBinaryBPP, EvolvedBinaryBCM, initCond, kick_info = Evolve.evolve(
            initialbinarytable=INIT_CONDITIONS_NO_BSE_COLUMNS, params=PARAMS_INI, randomseed=523574, n_per_block=100)

        pd.testing.assert_frame_equal(EvolvedBinaryBPP, BPP_DF, check_dtype=False, check_exact=False)
        pd.testing.assert_frame_equal(EvolvedBinaryBCM, BCM_DF, check_dtype=False, check_exact=False)

    def test_multi_evolve_with_dict_and_table(self):
        EvolvedBinaryBPP, EvolvedBinaryBCM, initCond, kick_info = Evolve.evolve(
            initialbinarytable=INIT_CONDITIONS, BSEDict=BSEDict, randomseed=523574, n_per_block=100)

        pd.testing.assert_frame_equal(EvolvedBinaryBPP, BPP_DF, check_dtype=False, check_exact=False)
        pd.testing.assert_frame_equal(EvolvedBinaryBCM, BCM_DF, check_dtype=False, check_exact=False)



# -*- coding: utf-8 -*-
# Copyright (C) Katelyn Breivik (2017 - 2021)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`plotting`
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.patches as mpatches

from matplotlib.lines import Line2D
from .utils import a_from_p, calc_Roche_radius
from .evolve import Evolve

__author__ = "Jeff Andrews <andrews.jeff@gmail.com>"
__all__ = [
    "evolve_binary",
    "plot_k_type",
    "plot_radius",
    "plot_mass",
    "plot_Teff",
    "plot_Mdot",
    "plot_P_orb",
    "plot_ecc",
    "plot_HR_diagram",
    "plot_binary_evol",
    "evolve_and_plot",
]

rsun_in_au = 215.0954
day_in_year = 365.242

# Colors
primary_color = "C0"
secondary_color = "C1"


def evolve_binary(initC, t_min=None, t_max=None, BSEDict={}):
    """
    Evolves a single binary with all timesteps written
    to the bcm array for plotting

    Parameters
    ----------
    initC : `pandas.DataFrame`
        initial conditions for binary to evolve

    t_min : `float`
        starting time for plot in Myr

    t_max : `float`
        ending time for plot in Myr

    BSEDict : `Dict`
        Dictionary containing all BSE flags needed

    Returns
    -------
    bcm : `pandas.DataFrame`
        binary evolution data form BSE's bcm dict
    """

    # Disable chained warning for now which throws
    # a warning for setting dtp and tphysf
    pd.options.mode.chained_assignment = None

    # Get highest BSE temporal resolution
    BSEDict["dtp"] = 0.01

    # To deal with the limited size of the bcm array, we need to reduce
    # the time resolution if we are evolving a binary for more than 100 Myr
    if t_max is not None:
        if t_min is not None:
            if t_max - t_min > 100:
                BSEDict["dtp"] = (t_max - t_min) / 10000
        else:
            if t_max > 100:
                BSEDict["dtp"] = t_max / 10000

    # Set maximum time for evolution
    if t_max is not None:
        initC["tphysf"] = t_max

    # Call evolution scripts
    bpp, bcm, initC, kick_info = Evolve.evolve(
        initialbinarytable=initC, BSEDict=BSEDict
    )

    # In case minimum and maximum times are not supplied by user
    if t_min is None:
        # BUG HERE - BSE sets first value to weird float
        # HACKED FIX
        t_phys = np.array(bcm.tphys)
        t_phys[0] = 0.0
        bcm.at[0, "tphys"] = t_phys
        t_min = bcm.tphys.min()
    if t_max is None:
        t_max = bcm.tphys.max()

    # Select for temporal range
    bcm = bcm.loc[(bcm.tphys <= t_max) & (bcm.tphys >= t_min)]

    return bcm


def plot_k_type(ax_1, ax_2, ax_k_type_list, times, k1_out, k2_out):
    """Plots the stellar types as a function of time

    Parameters
    ----------
    ax_1 : `matplotlib.Axes`
        Axis instance for kstar 1

    ax_2 : `matplotlib.Axes`
        Axis instance for kstar 2

    ax_k_type_list : `list`
        List of ktypes for the legend of the ktype bar plots

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    k1_out : `pandas.Series`
        Series of kstar 1 type at each binary evolution time step in Myr

    k2_out : `pandas.Series`
        Series of kstar 2 type at each binary evolution time step in Myr

    Returns
    -------
    Null
    """

    k_type_colors = [
        "plum",
        "sandybrown",
        "lightseagreen",
        "moccasin",
        "chartreuse",
        "deepskyblue",
        "gold",
        "rosybrown",
        "m",
        "darkgreen",
        "grey",
        "sienna",
        "palevioletred",
        "navy",
        "tan",
        "black",
    ]
    k_type = [
        "MS conv",
        "MS",
        "HG",
        "GB",
        "CHeB",
        "EAGB",
        "TPAGB",
        "HeMS",
        "HeHG",
        "HeGB",
        "HeWD",
        "COWD",
        "ONeWD",
        "NS",
        "BH",
        "no remnant",
    ]

    # k-type plots
    for a in [ax_1, ax_2]:

        a.axis("off")

        for j, k in enumerate(np.unique(k1_out)):
            a.fill_between(times[k1_out == k], 0.7, 0.9, color=k_type_colors[k])
        for j, k in enumerate(np.unique(k2_out)):
            a.fill_between(times[k2_out == k], 0.4, 0.6, color=k_type_colors[k])

        a.set_title("Stellar Type")

        a.set_xlim(np.min(times), np.max(times))

    # Add legend
    ax_k_type_list.axis("off")
    k_type_all = np.unique(np.array([k1_out, k2_out]))
    patches = [
        mpatches.Patch(color=k_type_colors[k], label=k_type[k]) for k in k_type_all
    ]
    leg = ax_k_type_list.legend(handles=patches, mode="expand", ncol=len(k_type_all))
    leg.get_frame().set_alpha(0.0)


def plot_radius(
    ax, times, R1_out, R2_out, M1_out, M2_out, P_orb_out, ecc_out, sys_obs, legend=True
):
    """Plots the stellar radii as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    R1_out : `pandas.Series`
        Series of primary radius at each binary evolution time step in Rsun

    R2_out : `pandas.Series`
        Series of secondary radius at each binary evolution time step in Rsun

    M1_out : `pandas.Series`
        Series of primary mass at each binary evolution time step in Msun

    M2_out : `pandas.Series`
        Series of secondary mass at each binary evolution time step in Msun

    P_orb_out : `pandas.Series`
        Series of orbital periods at each binary evolution time step in days

    ecc_out : `pandas.Series`
        Series of eccentricities at each binary evolution time step

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """
    # Radius
    ax.plot(times, R1_out, color=primary_color)
    ax.plot(times, R2_out, color=secondary_color)

    # Plot Roche radii - at periastron
    A_out = a_from_p(M1_out, M2_out, P_orb_out)  # in Rsun
    R1_Roche = calc_Roche_radius(M1_out, M2_out, A_out * (1.0 - ecc_out))  # in Rsun
    R2_Roche = calc_Roche_radius(M2_out, M1_out, A_out * (1.0 - ecc_out))  # in Rsun
    ax.plot(times, R1_Roche, color=primary_color, linestyle="--")
    ax.plot(times, R2_Roche, color=secondary_color, linestyle="--")

    for key, value in sys_obs.items():
        if key == "R1":
            ax.axhline(value, color=primary_color, linestyle="dashed")
        if key == "R2":
            ax.axhline(value, color=secondary_color, linestyle="dashed")

    ax.set_yscale("log")
    ax.set_ylim(0.5, ax.get_ylim()[1])

    ax.set_ylabel(r"Radius ($R_{\odot}$)")

    ax.set_xlim(np.min(times), np.max(times))

    if legend:
        custom_lines = [
            Line2D([0], [0], color="k", linestyle="solid"),
            Line2D([0], [0], color="k", linestyle="dashed"),
        ]
        ax.legend(custom_lines, ["Stellar Radius", "Roche Radius"], frameon=False)


def plot_mass(ax, times, M1_out, M2_out, sys_obs):
    """Plots the stellar mass as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    M1_out : `pandas.Series`
        Series of primary mass at each binary evolution time step in Msun

    M2_out : `pandas.Series`
        Series of secondary mass at each binary evolution time step in Msun

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """

    # Mass
    ax.plot(times, M1_out, color=primary_color)
    ax.plot(times, M2_out, color=secondary_color)

    for key, value in sys_obs.items():
        if key == "M1":
            ax.axhline(value, color=primary_color, linestyle="dashed")
        if key == "M2":
            ax.axhline(value, color=secondary_color, linestyle="dashed")

    ax.set_ylabel(r"Mass ($M_{\odot}$)")
    ax.set_xlim(np.min(times), np.max(times))


def plot_Teff(ax, times, Teff1_out, Teff2_out, sys_obs):
    """Plots the stellar effectvie temperature as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    Teff1_out : `pandas.Series`
        Series of primary effective temperature at each binary evolution time step in K

    Teff2_out : `pandas.Series`
        Series of secondary effective temperature at each binary evolution time step in K

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """

    # Effective temperature
    ax.plot(times, Teff1_out, color=primary_color)
    ax.plot(times, Teff2_out, color=secondary_color)

    for key, value in sys_obs.items():
        if key == "T1":
            ax.axhline(value, color=primary_color, linestyle="dashed")
        if key == "T2":
            ax.axhline(value, color=secondary_color, linestyle="dashed")

    ax.set_yscale("log")
    ax.set_ylabel(r"T$_{\rm eff}$ (K)")
    ax.set_xlim(np.min(times), np.max(times))
    ax.set_xlabel("Time (Myr)")


def plot_Mdot(ax, times, Mdot1_out, Mdot2_out, legend=True):
    """Plots the stellar effectvie temperature as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    Mdot1_out : `pandas.Series`
        Series of primary mass transfer rate at each binary evolution time step in Msun/yr

    Mdot2_out : `pandas.Series`
        Series of secondary mass transfer rate at each binary evolution time step in Msun/yr

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """

    # Mass accretion rate
    ax.plot(times, np.log10(np.clip(Mdot1_out, 1.0e-16, None)), color=primary_color)
    ax.plot(times, np.log10(np.clip(Mdot2_out, 1.0e-16, None)), color=secondary_color)
    ax.set_ylabel(r"Mass Accretion Rate ($M_{\odot}$ yr$^{-1}$)")

    ax.set_ylim(-14, ax.get_ylim()[1])
    ax.set_xlim(np.min(times), np.max(times))

    if legend:
        custom_lines = [
            Line2D([0], [0], color="C0", linestyle="solid"),
            Line2D([0], [0], color="C1", linestyle="solid"),
        ]
        ax.legend(custom_lines, ["Primary", "Secondary"], loc=2, frameon=False)


def plot_P_orb(ax, times, P_orb_out, t_max, sys_obs):
    """Plots the orbital period as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    P_orb_out : `pandas.Series`
        Series of orbital periods at each binary evolution time step in days

    t_max : `float`
        ending time for plot in Myr

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """

    ax.plot(times, P_orb_out, color="k")

    for key, value in sys_obs.items():
        if key == "P_orb":
            ax.axhline(value, color="k", linestyle="dashed")

    ax.set_xlim(np.min(times), t_max)

    ax.set_ylabel(r"P$_{\rm orb}$ (days)")
    ax.set_yscale("log")


def plot_ecc(ax, times, ecc_out, t_max, sys_obs):
    """Plots the eccentricity as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    ecc_out : `pandas.Series`
        Series of eccentricities at each binary evolution time step

    t_max : `float`
        ending time for plot in Myr

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """

    ax.plot(times, ecc_out, color="k")

    for key, value in sys_obs.items():
        if key == "ecc":
            ax.axhline(value, color="k", linestyle="dashed")

    ax.set_xlim(np.min(times), t_max)

    ax.set_ylabel("Eccentricity")
    ax.set_xlabel("Time (Myr)")


def plot_HR_diagram(ax, L1_out, L2_out, Teff1_out, Teff2_out):
    """Plots the eccentricity as a function of time

    Parameters
    ----------
    ax : `matplotlib.Axes`
        Axis instance for the plot

    times : `pandas.Series`
        Series of times at each binary evolution time step in Myr

    L1_out : `pandas.Series`
        Series of primary luminosities at each binary evolution time step in Lsun

    L2_out : `pandas.Series`
        Series of secondary luminosities at each binary evolution time step in Lsun

    Teff1_out : `pandas.Series`
        Series of primary effective temperature at each binary evolution time step in K

    Teff2_out : `pandas.Series`
        Series of secondary effective temperature at each binary evolution time step in K

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    Null
    """

    ax.plot(Teff1_out, L1_out, color=primary_color)
    ax.plot(Teff2_out, L2_out, color=secondary_color)

    ax.set_xscale("log")
    ax.set_yscale("log")
    ax.set_xlim(40000, 2000)

    ax.set_xlabel(r"log T$_{\rm eff}$ (K)")
    ax.set_ylabel("log L (erg/s)")

    ax.set_xticks([2000, 5000, 10000, 20000, 40000])


def plot_binary_evol(bcm, sys_obs={}):
    """Plots the full set of plots and kstar bar plots

    Parameters
    ----------
    bcm : `pandas.DataFrame`
        binary evolution data form BSE's bcm dict

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    fig : `matplotlib.figure`
        Figure containing all the plots!
    """

    fig, ax = plt.subplots(5, 1, figsize=(10, 10))

    gs = gridspec.GridSpec(4, 2, width_ratios=[2, 2], height_ratios=[1, 2, 2, 2])

    ax = np.array([plt.subplot(x) for x in gs])

    # k-type panels
    ax_k_type_list = fig.add_axes([0.08, 0.75, 0.9, 0.1])
    plot_k_type(
        ax[0],
        ax[1],
        ax_k_type_list,
        bcm.tphys,
        bcm.kstar_1.astype(int),
        bcm.kstar_2.astype(int),
    )

    # Radius panel
    plot_radius(
        ax[2],
        bcm.tphys,
        bcm.rad_1,
        bcm.rad_2,
        bcm.mass_1,
        bcm.mass_2,
        bcm.porb,
        bcm.ecc,
        sys_obs,
    )

    # Mass panel
    plot_mass(ax[4], bcm.tphys, bcm.mass_1, bcm.mass_2, sys_obs)

    # Teff panel
    plot_Teff(ax[6], bcm.tphys, bcm.teff_1, bcm.teff_2, sys_obs)

    # Mass accretion rate panel
    plot_Mdot(ax[3], bcm.tphys, bcm.deltam_1, bcm.deltam_2)

    # Orbital period panel
    plot_P_orb(
        ax[5],
        bcm.loc[bcm.kstar_2 < 15].tphys,
        bcm.loc[bcm.kstar_2 < 15].porb,
        np.max(bcm.tphys),
        sys_obs,
    )

    # Plot eccentricity panel
    plot_ecc(
        ax[7],
        bcm.loc[bcm.kstar_2 < 15].tphys,
        bcm.loc[bcm.kstar_2 < 15].ecc,
        np.max(bcm.tphys),
        sys_obs,
    )

    # Plot HR diagram
    # idx_1 = np.where(k1_out < 10)[0]
    # idx_2 = np.where(k2_out < 10)[0]
    # plot_HR_diagram(ax[7], L1_out[k1_out<10], L2_out[k2_out<10], Teff1_out[k1_out<10], Teff2_out[k2_out<10])

    # make the labels look nice
    gs.tight_layout(fig)

    return fig


def evolve_and_plot(initC, t_min=None, t_max=None, BSEDict=None, sys_obs={}):
    """
    Evolve and plot binaries as a function of time

    Parameters
    ----------
    initC : `pandas.DataFrame`
        initial conditions for binary to evolve

    t_min : `float or list`
        starting time for plot in Myr

    t_max : `float or list`
        ending time for plot in Myr

    BSEDict : `Dict`
        Dictionary containing all BSE flags needed

    sys_obs : `Dict`
        Dictionary containing keys for binary parameters with values to plot
        as vertical lines for each stellar component

    Returns
    -------
    all_figs : `list`
        list of all figures created
    """

    # Throw an error if user is plotting more than 10 binaries
    if len(initC) > 10:
        raise ValueError(
            "You have asked to plot more than 10 separate binaries. This could cause problems..."
        )

    # Iterate over all binaries in initC
    all_figs = []
    for i in range(len(initC)):

        # Check if t_min and t_max are lists
        if isinstance(t_min, list):
            t_min_tmp = t_min[i]
        else:
            t_min_tmp = t_min
        if isinstance(t_max, list):
            t_max_tmp = t_max[i]
        else:
            t_max_tmp = t_max

        # Evolve binary
        bcm = evolve_binary(
            initC.iloc[i: i + 1], t_min=t_min_tmp, t_max=t_max_tmp, BSEDict=BSEDict
        )

        # Plot binary
        fig = plot_binary_evol(bcm, sys_obs=sys_obs)

        # Add to list of figs
        all_figs.append(fig)

    # Return list of figs
    return all_figs



# -*- coding: utf-8 -*-
# Copyright (C) Carl Rodriguez (2020 - 2021)
#
# This file is part of cosmic.
#
# cosmic is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# cosmic is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with cosmic.  If not, see <http://www.gnu.org/licenses/>.

"""`King`
"""

import numpy as np
from numpy.random import uniform, normal
from scipy.integrate import RK45, quad, simps, cumtrapz
from scipy.interpolate import interp1d
from scipy.special import erf

__author__ = "Carl Rodriguez <carllouisrodriguez@gmail.com>"
__credits__ = "Carl Rodriguez <carllouisrodriguez@gmail.com>"
__all__ = ["calc_rho", "integrate_king_profile", "virial_radius_numerical", "find_sigma_sqr", "get_positions", "get_velocities", "scale_pos_and_vel", "draw_r_vr_vt"]

def calc_rho(w):
    """
    Returns the density (unnormalized) given w = psi/sigma^2
    Note that w(0) = w_0, the main parameter of a King profile
    """
    rho = np.exp(w)*erf(np.sqrt(w)) - np.sqrt(4.0*w/np.pi)*(1.0+2.0*w/3.0)
    return rho

def integrate_king_profile(w0,tidal_boundary=1e-6):
    """ 
    Integrate a King Profile of a given w_0 until the density (or phi/sigma^2) 
    drops below tidal_boundary limit (1e-8 times the central density by default)
    
    Let's define some things:
    The King potential is often expressed in terms of w = psi / sigma^2 (psi is phi0 - phi, so just positive potential)
    (note that sigma is the central velocity dispersion with an infinitely deep potential, and close otherwise)
    
    in the center, w_0 = psi_0 / sigma^2 (the free parameter of the King profile)
    
    The core radius is defined (King 1966) as r_c = sqrt(9 * sigma^2 / (4 pi G rho_0))
    
    If we define new scaled quantities
            r_tilda = r/r_c
            rho_tilda = rho/rho_o
    We can rewrite Poisson's equation, (1/r^2) d/dr (r^2 dphi/dr)  =  4 pi G rho as:
            d^2(r_tilda w)/dr_tilda^2 = 9 r_tilda rho_tilda
    
    After that, all we need is initial conditions:
    w(0) = w_0
    w'(0) = 0

    returns (radii, rho, phi, M_enclosed)
    """

    rho_0 = calc_rho(w0)

    def ode_rhs(r_tilda,w_vec):
        ## Unpack the w_vector (note I'm cheating by putting rho_0 as an element)
        w,w_dot,rho_0 = w_vec

        ## Compute rho and second derivative of w
        rho = calc_rho(w)
        w_ddot = -9*rho/rho_0 - 2*w_dot/r_tilda

        ## return dw/dr_tilda, d^2w/dr_tilda^2, d(rho_tilda)/dr_tilda = 0
        return (w_dot,w_ddot,0)

    king_profile = RK45(ode_rhs,1e-4,[w0,0,rho_0],1e5,rtol=1e-10,atol=1e-12)

    at_the_tidal_boundary = False
    r = []
    rho_r = []
    phi_r = []
    
    while not at_the_tidal_boundary:
        w,w_dot,rho_0 = king_profile.y
        rho = calc_rho(w)

        r.append(king_profile.t)
        rho_r.append(rho)
        phi_r.append(w)

        if rho/rho_0 < tidal_boundary:
            at_the_tidal_boundary = True
        else:
            king_profile.step()
            
    r = np.array(r)
    rho_r = np.array(rho_r)
    
    ## Finally compute the cumulative mass enclosed
    M_enclosed = cumtrapz(4*np.pi*r**2*rho_r, r, initial=0)
    
    return r, rho_r, phi_r, M_enclosed


def virial_radius_numerical(r, rho_r, M_enclosed):
    """
    Virial radius is best calculated directly.  Directly integrate 
    4*pi*r*rho*m_enclosed over the samples binding energy, then just divide 0.5 by that.
    """
    integral = simps(rho_r * r * 4 * np.pi * M_enclosed,r)

    return 1 / 2.0 / integral

def find_sigma_sqr(r_sample, r, rho_r, M_enclosed):
    """
    Find the 1D velocity dispersion at a given radius r using one of the
    spherial Jeans equations (and assuming velocity is isotropic)
    """

    ## This needs to be done continuously, so use interpolators
    rho_interp = interp1d(r,rho_r)
    M_enc_interp = interp1d(r,M_enclosed)
    
    jeans_integrand = lambda rr: rho_interp(rr) * M_enc_interp(rr) / rr / rr
    jeans_integrand = interp1d(r,rho_r*M_enclosed/r/r)

    integral, error = quad(jeans_integrand, r_sample, r[-1])
    return integral / rho_interp(r_sample)

def get_positions(N, r, M_enclosed):
    """
    This one's easy: the mass enclosed function is just the CDF of the mass
    density, so just invert that, and you've got positions.

    Note that here we've already normalized M_enclosed to 1 at r_tidal
    """

    # Use an interpolator, but flip x and y for the CDF
    interpolator = interp1d(M_enclosed, r, kind="cubic")

    X = uniform(size=N)

    positions = interpolator(X)

    return positions

def get_velocities(r, r_profile, psi_profile, M_enclosed_profile):
    """
    The correct way to generate velocities: start from the distribution function
    and use rejection sampling.

    returns (vr,vt) with same length as r
    """

    N = len(r)

    # create an interpolator for the psi
    psi_r = interp1d(r_profile,psi_profile)
    psi = psi_r(r)

    # Make an initial guess for every velocity (some fraction of the escape speed)
    x_rand = uniform(size=N)
    y_rand = uniform(size=N)

    resample = np.ones(N,dtype=bool)
    number_to_resample = N

    minF = (np.exp(psi) - 1)*2*psi
    v = x_rand * np.sqrt(2*psi)
    f_0 = y_rand*minF
    f = (np.exp(psi-v**2/2)-1)*v**2

    # Now do the rejection sampling
    while number_to_resample > 0: 
        v[resample] = x_rand[resample] * np.sqrt(2*psi[resample])
        f_0[resample] = y_rand[resample]*minF[resample]
        f[resample] = (np.exp(psi[resample]-v[resample]**2/2)-1)*v[resample]**2

        resample[resample] = f[resample] < f_0[resample]
        number_to_resample = np.sum(resample)
        x_rand[resample] = uniform(size=number_to_resample)
        y_rand[resample] = uniform(size=number_to_resample)

    theta = np.arccos(uniform(-1,1,N))

    vr = v*np.cos(theta)
    vt = v*np.sin(theta)

    return vr, vt


def scale_pos_and_vel(r,vr,vt):
    """
    Scale the positions and velocities to be in N-body units
    If we add binaries we'll do this again in initialcmctable.py

    takes r, vr, and vt as input

    returns (r,vr,vt) scaled to Henon units
    """

    # Normalize masses to 1/Mtotal
    mass = np.ones_like(r)/len(r)
    cumul_mass = np.cumsum(mass)

    radius = np.array(r)
    radius_p1 = np.append(radius[1:], [1e100])

    # Then compute the total kinetic and potential energy
    # There's probably a cleaner way to do the PE (this is a one-line version
    #  of the for loop we use in CMC; vectorized and pythonic, but sloppy)
    KE = 0.5 * sum(mass * (vr ** 2 + vt ** 2))
    PE = 0.5 * sum(
        mass[::-1]
        * np.cumsum((cumul_mass * (1.0 / radius - 1.0 / radius_p1))[::-1])
    )

    # Compute the position and velocity scalings
    rfac = 2 * PE
    vfac = 1.0 / np.sqrt(4 * KE)

    return r*rfac, vr*vfac, vt*vfac

def draw_r_vr_vt(N=100000, w_0=6, tidal_boundary=1e-6):
    """
    Draw random velocities and positions from the King profile.

    N              = number of stars
    w_0            = King concentration parameter (-psi/sigma^2) 
    tidal_boundary = ratio of rho/rho_0 where we truncate the tidal boundary

    returns (vr,vt,r) in G=M_cluster=1 units
    """
    # First integrate the differential equation of a King profile 
    r_profile, rho_profile, psi_profile, M_enclosed_profile = integrate_king_profile(w_0, tidal_boundary=tidal_boundary)

    ## Normalize the masses to 1 at r_tidal
    rho_profile /= M_enclosed_profile[-1]
    M_enclosed_profile /= M_enclosed_profile[-1]

    ## Normalize r (currently in units of the King core radius) to the virial radius
    r_profile /= virial_radius_numerical(r_profile, rho_profile, M_enclosed_profile)

    # Then draw the positions from the cumulative mass function
    r = get_positions(N, r_profile, M_enclosed_profile)

    # Sort the radii (needed later)
    r = np.sort(r)

    # Finally, draw the velocities using the radii and one of the Jeans
    # equations
    vr, vt = get_velocities(r, r_profile, psi_profile, M_enclosed_profile)

    # Scale into Henon units
    r, vr, vt = scale_pos_and_vel(r,vr,vt)

    return r, vr, vt
